# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from argparse import Namespace

from botocore.exceptions import ClientError

import awscli
from awscli.compat import ZIP_COMPRESSION_MODE, StringIO
from awscli.customizations.codedeploy.push import Push
from awscli.customizations.exceptions import ParamValidationError
from awscli.testutils import mock, unittest


class TestPush(unittest.TestCase):
    def setUp(self):
        self.application_name = 'MyApp'
        self.description = 'MyApp revision'
        self.source = '/tmp'
        self.appspec = 'appspec.yml'
        self.appspec_path = f'{self.source}/{self.appspec}'
        self.bucket = 'foo'
        self.key = 'bar/baz.zip'
        self.s3_location = 's3://' + self.bucket + '/' + self.key
        self.eTag = '"1a2b3cd45e"'
        self.version_id = '12341234-1234-1234-1234-123412341234'
        self.upload_id = 'upload_id'
        self.region = 'us-east-1'
        self.endpoint_url = 'https://codedeploy.aws.amazon.com'

        self.args = Namespace()
        self.args.application_name = self.application_name
        self.args.s3_location = self.s3_location
        self.args.ignore_hidden_files = False
        self.args.no_ignore_hidden_files = False
        self.args.description = self.description
        self.args.source = self.source

        self.globals = Namespace()
        self.globals.region = self.region
        self.globals.endpoint_url = self.endpoint_url
        self.globals.verify_ssl = False

        self.upload_response = {
            'ETag': self.eTag,
            'VersionId': self.version_id,
        }
        self.revision = {
            'revisionType': 'S3',
            's3Location': {
                'bucket': self.bucket,
                'key': self.key,
                'bundleType': 'zip',
                'eTag': self.eTag,
                'version': self.version_id,
            },
        }

        self.bundle_mock = mock.MagicMock()
        self.bundle_mock.tell.return_value = 5 << 20
        self.bundle_mock.read.return_value = b'a' * (5 << 20)
        self.bundle_mock.__enter__.return_value = self.bundle_mock
        self.bundle_mock.__exit__.return_value = None

        self.zipfile_mock = mock.MagicMock()
        self.zipfile_mock.write.return_value = None
        self.zipfile_mock.close.return_value = None
        self.zipfile_mock.__enter__.return_value = self.zipfile_mock
        self.zipfile_mock.__exit__.return_value = None

        self.session = mock.MagicMock()

        self.push = Push(self.session)
        self.push.s3 = mock.MagicMock()
        self.push.s3.put_object.return_value = self.upload_response
        self.push.s3.create_multipart_upload.return_value = {
            'UploadId': self.upload_id
        }
        self.push.s3.upload_part.return_value = {'ETag': self.eTag}
        self.push.s3.complete_multipart_upload.return_value = (
            self.upload_response
        )
        self.push.codedeploy = mock.MagicMock()

    def test_run_main_throws_on_invalid_args(self):
        self.push._validate_args = mock.MagicMock()
        self.push._validate_args.side_effect = RuntimeError()
        with self.assertRaises(RuntimeError):
            self.push._run_main(self.args, self.globals)

    def test_run_main_creates_clients(self):
        self.push._validate_args = mock.MagicMock()
        self.push._push = mock.MagicMock()
        self.push._run_main(self.args, self.globals)
        self.session.create_client.assert_has_calls(
            [
                mock.call(
                    'codedeploy',
                    region_name=self.region,
                    endpoint_url=self.endpoint_url,
                    verify=self.globals.verify_ssl,
                ),
                mock.call('s3', region_name=self.region),
            ]
        )

    def test_run_main_calls_push(self):
        self.push._validate_args = mock.MagicMock()
        self.push._push = mock.MagicMock()
        self.push._run_main(self.args, self.globals)
        self.push._push.assert_called_with(self.args)

    @mock.patch.object(
        awscli.customizations.codedeploy.push, 'validate_s3_location'
    )
    def test_validate_args_throws_on_invalid_s3_url(
        self, validate_s3_location
    ):
        self.args.s3_location = 's3:/foo/bar/baz'
        validate_s3_location.side_effect = RuntimeError()
        with self.assertRaises(RuntimeError):
            self.push._validate_args(self.args)

    def test_validate_args_throws_on_ignore_and_no_ignore_hidden_files(self):
        self.args.ignore_hidden_files = True
        self.args.no_ignore_hidden_files = True
        with self.assertRaises(ParamValidationError):
            self.push._validate_args(self.args)

    def test_validate_args_default_description(self):
        self.args.description = None
        self.push._validate_args(self.args)
        self.assertRegex(self.args.description, 'Uploaded by AWS CLI .* UTC')

    def test_push_throws_on_upload_to_s3_error(self):
        self.args.bucket = self.bucket
        self.args.key = self.key
        self.push._compress = mock.MagicMock(return_value=self.bundle_mock)
        self.push._upload_to_s3 = mock.MagicMock()
        self.push._upload_to_s3.side_effect = RuntimeError()
        with self.assertRaises(RuntimeError):
            self.push._push(self.args)

    def test_push_strips_quotes_from_etag(self):
        self.args.bucket = self.bucket
        self.args.key = self.key
        self.push._compress = mock.MagicMock(return_value=self.bundle_mock)
        self.push._upload_to_s3 = mock.MagicMock(
            return_value=self.upload_response
        )
        self.push._register_revision = mock.MagicMock()
        self.push._push(self.args)
        self.push._register_revision.assert_called_with(self.args)
        self.assertEqual(
            str(self.args.eTag), self.upload_response['ETag'].replace('"', "")
        )

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_push_output_message(self, stdout_mock):
        self.args.bucket = self.bucket
        self.args.key = self.key
        self.push._compress = mock.MagicMock(return_value=self.bundle_mock)
        self.push._upload_to_s3 = mock.MagicMock(
            return_value=self.upload_response
        )
        self.push._register_revision = mock.MagicMock()
        self.push._push(self.args)
        output = stdout_mock.getvalue().strip()
        expected_revision_output = (
            '--s3-location bucket={0},key={1},'
            'bundleType=zip,eTag={2},version={3}'.format(
                self.bucket,
                self.key,
                self.eTag.replace('"', ""),
                self.version_id,
            )
        )
        expected_output = (
            'To deploy with this revision, run:\n'
            'aws deploy create-deployment '
            f'--application-name {self.application_name} {expected_revision_output} '
            '--deployment-group-name <deployment-group-name> '
            '--deployment-config-name <deployment-config-name> '
            '--description <description>'
        )
        self.assertEqual(expected_output, output)

    @mock.patch('zipfile.ZipFile')
    @mock.patch('tempfile.TemporaryFile')
    @mock.patch('os.path')
    @mock.patch('os.walk')
    def test_compress_throws_when_no_appspec(self, walk, path, tf, zf):
        walk.return_value = [(self.source, [], ['noappspec.yml'])]
        noappsec_path = self.source + '/noappspec.yml'
        path.join.return_value = noappsec_path
        path.sep = '/'
        path.abspath.side_effect = [self.source, noappsec_path]
        tf.return_value = self.bundle_mock
        zf.return_value = self.zipfile_mock
        with self.assertRaises(RuntimeError):
            with self.push._compress(
                self.args.source, self.args.ignore_hidden_files
            ):
                pass

    @mock.patch('zipfile.ZipFile')
    @mock.patch('tempfile.TemporaryFile')
    @mock.patch('os.path')
    @mock.patch('os.walk')
    def test_compress_writes_to_zip_file(self, walk, path, tf, zf):
        walk.return_value = [(self.source, [], [self.appspec])]
        path.join.return_value = self.appspec_path
        path.sep = '/'
        path.abspath.side_effect = [self.source, self.appspec_path]
        tf.return_value = self.bundle_mock
        zf.return_value = self.zipfile_mock
        with self.push._compress(
            self.args.source, self.args.ignore_hidden_files
        ):
            zf.assert_called_with(mock.ANY, 'w', allowZip64=True)
            zf().write.assert_called_with(
                '/tmp/appspec.yml', self.appspec, ZIP_COMPRESSION_MODE
            )

    def test_upload_to_s3_with_put_object(self):
        self.args.bucket = self.bucket
        self.args.key = self.key
        response = self.push._upload_to_s3(self.args, self.bundle_mock)
        self.assertDictEqual(self.upload_response, response)
        self.push.s3.put_object.assert_called_with(
            Bucket=self.bucket, Key=self.key, Body=self.bundle_mock
        )
        self.assertFalse(self.push.s3.create_multipart_upload.called)
        self.assertFalse(self.push.s3.upload_part.called)
        self.assertFalse(self.push.s3.complete_multipart_upload.called)
        self.assertFalse(self.push.s3.abort_multipart_upload.called)

    def test_upload_to_s3_with_multipart_upload(self):
        self.args.bucket = self.bucket
        self.args.key = self.key
        self.bundle_mock.tell.return_value = 6 << 20
        self.bundle_mock.read.return_value = b'a' * (6 << 20)
        response = self.push._upload_to_s3(self.args, self.bundle_mock)
        self.assertDictEqual(self.upload_response, response)
        self.assertFalse(self.push.s3.put_object.called)
        self.push.s3.create_multipart_upload.assert_called_with(
            Bucket=self.bucket, Key=self.key
        )
        self.push.s3.upload_part.assert_called_with(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.upload_id,
            PartNumber=1,
            Body=mock.ANY,
        )
        self.push.s3.complete_multipart_upload.assert_called_with(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.upload_id,
            MultipartUpload={'Parts': [{'PartNumber': 1, 'ETag': self.eTag}]},
        )
        self.assertFalse(self.push.s3.abort_multipart_upload.called)

    def test_upload_to_s3_with_multipart_upload_aborted_on_error(self):
        self.args.bucket = self.bucket
        self.args.key = self.key
        self.bundle_mock.tell.return_value = 6 << 20
        self.bundle_mock.read.return_value = b'a' * (6 << 20)
        self.push.s3.upload_part.side_effect = ClientError(
            {'Error': {'Code': 'Error', 'Message': 'Error'}}, 'UploadPart'
        )
        with self.assertRaises(ClientError):
            self.push._upload_to_s3(self.args, self.bundle_mock)
        self.assertFalse(self.push.s3.put_object.called)
        self.push.s3.create_multipart_upload.assert_called_with(
            Bucket=self.bucket, Key=self.key
        )
        self.assertTrue(self.push.s3.upload_part.called)
        self.assertFalse(self.push.s3.complete_multipart_upload.called)
        self.push.s3.abort_multipart_upload.assert_called_with(
            Bucket=self.bucket, Key=self.key, UploadId=self.upload_id
        )

    def test_register_revision(self):
        self.args.bucket = self.bucket
        self.args.key = self.key
        self.args.eTag = self.eTag
        self.args.version = self.version_id
        self.push._register_revision(self.args)
        self.push.codedeploy.register_application_revision.assert_called_with(
            applicationName=self.application_name,
            description=self.description,
            revision=self.revision,
        )


if __name__ == "__main__":
    unittest.main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sqlite3
from unittest.mock import MagicMock, PropertyMock, patch

import pytest
from botocore.exceptions import MD5UnavailableError
from botocore.session import Session

from awscli.telemetry import (
    CLISessionData,
    CLISessionDatabaseConnection,
    CLISessionDatabaseReader,
    CLISessionDatabaseSweeper,
    CLISessionDatabaseWriter,
    CLISessionGenerator,
    CLISessionOrchestrator,
    add_session_id_component_to_user_agent_extra,
)
from tests.markers import skip_if_windows


@pytest.fixture
def session_conn():
    conn = CLISessionDatabaseConnection(
        connection=sqlite3.connect(
            # Use an in-memory db for testing.
            ':memory:',
            check_same_thread=False,
            isolation_level=None,
        ),
    )
    # Write an initial record.
    conn.execute(
        """
            INSERT OR REPLACE INTO session (
                key, session_id, timestamp
            ) VALUES ('first_key', 'first_id', 5555555555)
        """
    )
    # Overwrite host id with deterministic value for testing.
    conn.execute(
        """
            INSERT OR REPLACE INTO host_id (
                key, id
            ) VALUES (0, 'my-hostname')
        """
    )
    return conn


@pytest.fixture
def session_writer(session_conn):
    return CLISessionDatabaseWriter(session_conn)


@pytest.fixture
def session_reader(session_conn):
    return CLISessionDatabaseReader(session_conn)


@pytest.fixture
def session_sweeper(session_conn):
    return CLISessionDatabaseSweeper(session_conn)


@pytest.fixture
def session_generator():
    return CLISessionGenerator()


@pytest.fixture
def expired_data(session_writer, session_reader, session_sweeper):
    # Write an expired record.
    session_writer.write(
        CLISessionData(
            key='expired_key',
            session_id='expired_id',
            timestamp=1000000000,
        )
    )
    # Ensure expired record exists.
    assert session_reader.read('expired_key') is not None
    yield
    # Ensure cleanup after test is run.
    session_sweeper.sweep(1000000001)


class TestCLISessionDatabaseConnection:
    def test_ensure_database_setup(self, session_conn):
        cursor = session_conn.execute(
            """
                SELECT name
                FROM sqlite_master
                WHERE type='table';
            """
        )
        assert cursor.fetchall() == [('session',), ('host_id',)]

    def test_timeout_does_not_raise_exception(self, session_conn):
        test_query = """
            SELECT name
            FROM sqlite_master
            WHERE type='table'
            AND name='session';
        """

        class FakeConnection(sqlite3.Connection):
            def execute(self, query, *parameters):
                # Simulate timeout by always raising.
                if query == test_query:
                    raise sqlite3.OperationalError()
                # Mock host id count query.
                cur = MagicMock()
                cur.fetchone.return_value = (1,)
                return cur

        fake_conn = CLISessionDatabaseConnection(FakeConnection(":memory:"))
        cursor = fake_conn.execute(test_query)
        assert cursor.fetchall() == []


class TestCLISessionDatabaseWriter:
    def test_write(self, session_writer, session_reader, session_sweeper):
        session_writer.write(
            CLISessionData(
                key='new-key',
                session_id='new-id',
                timestamp=1000000000,
            )
        )
        session_data = session_reader.read('new-key')
        assert session_data.key == 'new-key'
        assert session_data.session_id == 'new-id'
        assert session_data.timestamp == 1000000000
        session_sweeper.sweep(1000000001)


class TestCLISessionDatabaseReader:
    def test_read(self, session_reader):
        session_data = session_reader.read('first_key')
        assert session_data.key == 'first_key'
        assert session_data.session_id == 'first_id'
        assert session_data.timestamp == 5555555555

    def test_read_nonexistent_record(self, session_reader):
        session_data = session_reader.read('bad_key')
        assert session_data is None

    def test_read_host_id(self, session_reader):
        host_id = session_reader.read_host_id()
        assert host_id == 'my-hostname'


class TestCLISessionDatabaseSweeper:
    def test_sweep(self, expired_data, session_reader, session_sweeper):
        session_sweeper.sweep(1000000001)
        swept_data = session_reader.read('expired_key')
        assert swept_data is None

    def test_sweep_not_expired(
        self, expired_data, session_reader, session_sweeper
    ):
        session_sweeper.sweep(1000000000)
        swept_data = session_reader.read('expired_key')
        assert swept_data is not None

    def test_sweep_never_raises(self, session_sweeper):
        # Normally this would raise `sqlite3.ProgrammingError`,
        # but the `sweep` method catches bare exceptions.
        session_sweeper.sweep({'bad': 'input'})


class TestCLISessionGenerator:
    def test_generate_session_id(self, session_generator):
        session_id = session_generator.generate_session_id(
            'my-hostname',
            'my-tty',
            1000000000,
        )
        assert session_id == 'd949713b13ee'

    def test_generate_cache_key(self, session_generator):
        cache_key = session_generator.generate_cache_key(
            'my-hostname',
            'my-tty',
        )
        assert cache_key == 'b1ca2be0ffac'

    @patch('awscli.telemetry.get_md5')
    def test_checksum_fips_fallback(self, patched_get_md5, session_generator):
        patched_get_md5.side_effect = MD5UnavailableError()
        session_id = session_generator.generate_session_id(
            'my-hostname',
            'my-tty',
            1000000000,
        )
        assert session_id == '183b154db015'


@skip_if_windows
@patch('sys.stdin')
@patch('time.time', return_value=5555555555)
@patch('os.ttyname', return_value='my-tty')
class TestCLISessionOrchestrator:
    def test_session_id_gets_cached(
        self,
        patched_tty_name,
        patched_time,
        patched_stdin,
        session_sweeper,
        session_generator,
        session_reader,
        session_writer,
    ):
        patched_stdin.fileno.return_value = None
        orchestrator = CLISessionOrchestrator(
            session_generator, session_writer, session_reader, session_sweeper
        )
        assert orchestrator.session_id == '881cea8546fa'

        session_data = session_reader.read(orchestrator.cache_key)
        assert session_data.key == orchestrator.cache_key
        assert session_data.session_id == orchestrator.session_id
        assert session_data.timestamp == 5555555555

    def test_cached_session_id_updated_if_expired(
        self,
        patched_tty_name,
        patched_time,
        patched_stdin,
        session_sweeper,
        session_generator,
        session_reader,
        session_writer,
    ):
        patched_stdin.fileno.return_value = None

        # First, generate and cache a session id.
        orchestrator_1 = CLISessionOrchestrator(
            session_generator, session_writer, session_reader, session_sweeper
        )
        session_id_1 = orchestrator_1.session_id
        session_data_1 = session_reader.read(orchestrator_1.cache_key)
        assert session_data_1.session_id == session_id_1

        # Update the timestamp and get the new session id.
        patched_time.return_value = 7777777777
        orchestrator_2 = CLISessionOrchestrator(
            session_generator, session_writer, session_reader, session_sweeper
        )
        session_id_2 = orchestrator_2.session_id
        session_data_2 = session_reader.read(orchestrator_2.cache_key)

        # Cache key should be the same.
        assert session_data_2.key == session_data_1.key
        # Session id and timestamp should be updated.
        assert session_data_2.session_id == session_id_2
        assert session_data_2.session_id != session_data_1.session_id
        assert session_data_2.timestamp == 7777777777
        assert session_data_2.timestamp != session_data_1.timestamp

    def test_cached_session_id_not_updated_if_valid(
        self,
        patched_tty_name,
        patched_time,
        patched_stdin,
        session_sweeper,
        session_generator,
        session_reader,
        session_writer,
    ):
        patched_stdin.fileno.return_value = None

        # First, generate and cache a session id.
        orchestrator_1 = CLISessionOrchestrator(
            session_generator, session_writer, session_reader, session_sweeper
        )
        session_id_1 = orchestrator_1.session_id
        session_data_1 = session_reader.read(orchestrator_1.cache_key)
        assert session_data_1.session_id == session_id_1

        # Update the timestamp.
        patched_time.return_value = 5555555556
        orchestrator_2 = CLISessionOrchestrator(
            session_generator, session_writer, session_reader, session_sweeper
        )
        session_id_2 = orchestrator_2.session_id
        session_data_2 = session_reader.read(orchestrator_2.cache_key)

        # Cache key should be the same.
        assert session_data_2.key == session_data_1.key
        # Session id should not be updated.
        assert session_data_2.session_id == session_id_2
        assert session_data_2.session_id == session_data_1.session_id
        # Only timestamp should be updated.
        assert session_data_2.timestamp == 5555555556
        assert session_data_2.timestamp != session_data_1.timestamp


def test_add_session_id_component_to_user_agent_extra():
    session = MagicMock(Session)
    session.user_agent_extra = ''
    orchestrator = MagicMock(CLISessionOrchestrator)
    orchestrator.session_id = 'my-session-id'
    add_session_id_component_to_user_agent_extra(session, orchestrator)
    assert session.user_agent_extra == 'sid/my-session-id'


def test_entrypoint_catches_bare_exceptions():
    mock_orchestrator = MagicMock(CLISessionOrchestrator)
    type(mock_orchestrator).session_id = PropertyMock(side_effect=Exception)
    session = MagicMock(Session)
    add_session_id_component_to_user_agent_extra(session, mock_orchestrator)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

from botocore.exceptions import ClientError
from s3transfer.manager import TransferManager

from awscli.compat import queue
from awscli.customizations.s3.fileinfo import FileInfo
from awscli.customizations.s3.results import (
    CommandResultRecorder,
    DoneResultSubscriber,
    DryRunResult,
    FailureResult,
    ProgressResultSubscriber,
    QueuedResult,
    QueuedResultSubscriber,
    ResultProcessor,
    ResultRecorder,
    SuccessResult,
)
from awscli.customizations.s3.s3handler import (
    CopyRequestSubmitter,
    DeleteRequestSubmitter,
    DownloadRequestSubmitter,
    DownloadStreamRequestSubmitter,
    LocalDeleteRequestSubmitter,
    S3TransferHandler,
    S3TransferHandlerFactory,
    UploadRequestSubmitter,
    UploadStreamRequestSubmitter,
)
from awscli.customizations.s3.subscribers import (
    DeleteSourceFileSubscriber,
    DeleteSourceObjectSubscriber,
    DirectoryCreatorSubscriber,
    ProvideETagSubscriber,
    ProvideFullObjectChecksumSubscriber,
    ProvideLastModifiedTimeSubscriber,
    ProvideSizeSubscriber,
    ProvideUploadContentTypeSubscriber,
    SetMetadataDirectivePropsSubscriber,
    SetTagsSubscriber,
)
from awscli.customizations.s3.transferconfig import RuntimeConfig
from awscli.customizations.s3.utils import (
    MAX_UPLOAD_SIZE,
    NonSeekableStream,
    StdoutBytesWriter,
    WarningResult,
)
from awscli.testutils import FileCreator, mock, unittest


def runtime_config(**kwargs):
    return RuntimeConfig().build_config(**kwargs)


class TestS3TransferHandlerFactory(unittest.TestCase):
    def setUp(self):
        self.cli_params = {}
        self.runtime_config = runtime_config()
        self.transfer_manager = mock.Mock()
        self.result_queue = queue.Queue()

    def test_call(self):
        factory = S3TransferHandlerFactory(self.cli_params)
        self.assertIsInstance(
            factory(self.transfer_manager, self.result_queue),
            S3TransferHandler,
        )


class TestS3TransferHandler(unittest.TestCase):
    def setUp(self):
        self.result_queue = queue.Queue()
        self.result_recorder = ResultRecorder()
        self.processed_results = []
        self.result_processor = ResultProcessor(
            self.result_queue,
            [self.result_recorder, self.processed_results.append],
        )
        self.command_result_recorder = CommandResultRecorder(
            self.result_queue, self.result_recorder, self.result_processor
        )

        self.transfer_manager = mock.Mock(spec=TransferManager)
        self.transfer_manager.__enter__ = mock.Mock()
        self.transfer_manager.__exit__ = mock.Mock()
        self.parameters = {}
        self.s3_transfer_handler = S3TransferHandler(
            self.transfer_manager,
            self.parameters,
            self.command_result_recorder,
        )

    def test_call_return_command_result(self):
        num_failures = 5
        num_warnings = 3
        self.result_recorder.files_failed = num_failures
        self.result_recorder.files_warned = num_warnings
        command_result = self.s3_transfer_handler.call([])
        self.assertEqual(command_result, (num_failures, num_warnings))

    def test_enqueue_uploads(self):
        fileinfos = []
        num_transfers = 5
        for _ in range(num_transfers):
            fileinfos.append(
                FileInfo(
                    src='filename', dest='bucket/key', operation_name='upload'
                )
            )

        self.s3_transfer_handler.call(fileinfos)
        self.assertEqual(
            self.transfer_manager.upload.call_count, num_transfers
        )

    def test_enqueue_downloads(self):
        fileinfos = []
        num_transfers = 5
        for _ in range(num_transfers):
            fileinfos.append(
                FileInfo(
                    src='bucket/key',
                    dest='filename',
                    compare_key='key',
                    operation_name='download',
                )
            )

        self.s3_transfer_handler.call(fileinfos)
        self.assertEqual(
            self.transfer_manager.download.call_count, num_transfers
        )

    def test_enqueue_copies(self):
        fileinfos = []
        num_transfers = 5
        for _ in range(num_transfers):
            fileinfos.append(
                FileInfo(
                    src='sourcebucket/sourcekey',
                    dest='bucket/key',
                    compare_key='key',
                    operation_name='copy',
                )
            )

        self.s3_transfer_handler.call(fileinfos)
        self.assertEqual(self.transfer_manager.copy.call_count, num_transfers)

    def test_exception_when_enqueuing(self):
        fileinfos = [
            FileInfo(
                src='filename', dest='bucket/key', operation_name='upload'
            )
        ]
        self.transfer_manager.__exit__.side_effect = Exception(
            'some exception'
        )
        command_result = self.s3_transfer_handler.call(fileinfos)
        # Exception should have been raised casing the command result to
        # have failed results of one.
        self.assertEqual(command_result, (1, 0))

    def test_enqueue_upload_stream(self):
        self.parameters['is_stream'] = True
        self.s3_transfer_handler.call(
            [FileInfo(src='-', dest='bucket/key', operation_name='upload')]
        )
        self.assertEqual(self.transfer_manager.upload.call_count, 1)
        upload_call_kwargs = self.transfer_manager.upload.call_args[1]
        self.assertIsInstance(upload_call_kwargs['fileobj'], NonSeekableStream)

    def test_enqueue_dowload_stream(self):
        self.parameters['is_stream'] = True
        self.s3_transfer_handler.call(
            [
                FileInfo(
                    src='bucket/key',
                    dest='-',
                    compare_key='key',
                    operation_name='download',
                )
            ]
        )
        self.assertEqual(self.transfer_manager.download.call_count, 1)
        download_call_kwargs = self.transfer_manager.download.call_args[1]
        self.assertIsInstance(
            download_call_kwargs['fileobj'], StdoutBytesWriter
        )

    def test_enqueue_deletes(self):
        fileinfos = []
        num_transfers = 5
        for _ in range(num_transfers):
            fileinfos.append(
                FileInfo(
                    src='bucket/key',
                    dest=None,
                    operation_name='delete',
                    src_type='s3',
                )
            )

        self.s3_transfer_handler.call(fileinfos)
        self.assertEqual(
            self.transfer_manager.delete.call_count, num_transfers
        )

    def test_enqueue_local_deletes(self):
        fileinfos = []
        num_transfers = 5
        for _ in range(num_transfers):
            fileinfos.append(
                FileInfo(
                    src='myfile',
                    dest=None,
                    operation_name='delete',
                    src_type='local',
                )
            )

        self.s3_transfer_handler.call(fileinfos)
        # The number of processed results will be equal to:
        # number_of_local_deletes * 2 + 1
        # The 2 represents the QueuedResult and SuccessResult/FailureResult
        # for each transfer
        # The 1 represents the TotalFinalSubmissionResult
        self.assertEqual(len(self.processed_results), 11)

        # Make sure that the results are as expected by checking just one
        # of them
        first_submitted_result = self.processed_results[0]
        self.assertEqual(first_submitted_result.transfer_type, 'delete')
        self.assertTrue(first_submitted_result.src.endswith('myfile'))

        # Also make sure that transfer manager's delete() was never called
        self.assertEqual(self.transfer_manager.delete.call_count, 0)

    def test_notifies_total_submissions(self):
        fileinfos = []
        num_transfers = 5
        for _ in range(num_transfers):
            fileinfos.append(
                FileInfo(
                    src='bucket/key',
                    dest='filename',
                    compare_key='key',
                    operation_name='download',
                )
            )

        self.s3_transfer_handler.call(fileinfos)
        self.assertEqual(
            self.result_recorder.final_expected_files_transferred,
            num_transfers,
        )

    def test_notifies_total_submissions_accounts_for_skips(self):
        fileinfos = []
        num_transfers = 5
        for _ in range(num_transfers):
            fileinfos.append(
                FileInfo(
                    src='bucket/key',
                    dest='filename',
                    compare_key='key',
                    operation_name='download',
                )
            )

        # Add a fileinfo that should get skipped. To skip, we do a glacier
        # download.
        fileinfos.append(
            FileInfo(
                src='bucket/key',
                dest='filename',
                operation_name='download',
                compare_key='key',
                associated_response_data={'StorageClass': 'GLACIER'},
            )
        )
        self.s3_transfer_handler.call(fileinfos)
        # Since the last glacier download was skipped the final expected
        # total should be equal to the number of transfers provided in the
        # for loop.
        self.assertEqual(
            self.result_recorder.final_expected_files_transferred,
            num_transfers,
        )


class BaseTransferRequestSubmitterTest(unittest.TestCase):
    def setUp(self):
        self.transfer_manager = mock.Mock(spec=TransferManager)
        self.result_queue = queue.Queue()
        self.cli_params = {}
        self.filename = 'myfile'
        self.bucket = 'mybucket'
        self.key = 'mykey'


class TestUploadRequestSubmitter(BaseTransferRequestSubmitterTest):
    def setUp(self):
        super(TestUploadRequestSubmitter, self).setUp()
        self.transfer_request_submitter = UploadRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )

    def test_can_submit(self):
        fileinfo = FileInfo(
            src=self.filename,
            dest=self.bucket + '/' + self.key,
            operation_name='upload',
        )
        self.assertTrue(self.transfer_request_submitter.can_submit(fileinfo))
        fileinfo.operation_name = 'foo'
        self.assertFalse(self.transfer_request_submitter.can_submit(fileinfo))

    def test_submit(self):
        fileinfo = FileInfo(
            src=self.filename, dest=self.bucket + '/' + self.key
        )
        self.cli_params['guess_mime_type'] = True  # Default settings
        future = self.transfer_request_submitter.submit(fileinfo)

        self.assertIs(self.transfer_manager.upload.return_value, future)
        upload_call_kwargs = self.transfer_manager.upload.call_args[1]
        self.assertEqual(upload_call_kwargs['fileobj'], self.filename)
        self.assertEqual(upload_call_kwargs['bucket'], self.bucket)
        self.assertEqual(upload_call_kwargs['key'], self.key)
        self.assertEqual(upload_call_kwargs['extra_args'], {})

        # Make sure the subscriber applied are of the correct type and order
        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            ProvideUploadContentTypeSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = upload_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_submit_with_extra_args(self):
        fileinfo = FileInfo(
            src=self.filename, dest=self.bucket + '/' + self.key
        )
        # Set some extra argument like storage_class to make sure cli
        # params get mapped to request parameters.
        self.cli_params['storage_class'] = 'STANDARD_IA'
        self.transfer_request_submitter.submit(fileinfo)

        upload_call_kwargs = self.transfer_manager.upload.call_args[1]
        self.assertEqual(
            upload_call_kwargs['extra_args'], {'StorageClass': 'STANDARD_IA'}
        )

    def test_submit_when_content_type_specified(self):
        fileinfo = FileInfo(
            src=self.filename, dest=self.bucket + '/' + self.key
        )
        self.cli_params['content_type'] = 'text/plain'
        self.transfer_request_submitter.submit(fileinfo)

        upload_call_kwargs = self.transfer_manager.upload.call_args[1]
        self.assertEqual(
            upload_call_kwargs['extra_args'], {'ContentType': 'text/plain'}
        )
        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = upload_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_submit_when_no_guess_content_mime_type(self):
        fileinfo = FileInfo(
            src=self.filename, dest=self.bucket + '/' + self.key
        )
        self.cli_params['guess_mime_type'] = False
        self.transfer_request_submitter.submit(fileinfo)

        upload_call_kwargs = self.transfer_manager.upload.call_args[1]
        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = upload_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_warn_on_too_large_transfer(self):
        fileinfo = FileInfo(
            src=self.filename,
            dest=self.bucket + '/' + self.key,
            size=MAX_UPLOAD_SIZE + 1,
        )
        future = self.transfer_request_submitter.submit(fileinfo)

        # A warning should have been submitted because it is too large.
        warning_result = self.result_queue.get()
        self.assertIsInstance(warning_result, WarningResult)
        self.assertIn('exceeds s3 upload limit', warning_result.message)

        # Make sure that the transfer was still attempted
        self.assertIs(self.transfer_manager.upload.return_value, future)
        self.assertEqual(len(self.transfer_manager.upload.call_args_list), 1)

    def test_dry_run(self):
        self.cli_params['dryrun'] = True
        self.transfer_request_submitter = UploadRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )
        fileinfo = FileInfo(
            src=self.filename,
            src_type='local',
            operation_name='upload',
            dest=self.bucket + '/' + self.key,
            dest_type='s3',
        )
        self.transfer_request_submitter.submit(fileinfo)

        result = self.result_queue.get()
        self.assertIsInstance(result, DryRunResult)
        self.assertEqual(result.transfer_type, 'upload')
        self.assertTrue(result.src.endswith(self.filename))
        self.assertEqual(result.dest, 's3://' + self.bucket + '/' + self.key)

    def test_submit_move_adds_delete_source_subscriber(self):
        fileinfo = FileInfo(
            src=self.filename, dest=self.bucket + '/' + self.key
        )
        self.cli_params['guess_mime_type'] = True  # Default settings
        self.cli_params['is_move'] = True
        self.transfer_request_submitter.submit(fileinfo)
        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            ProvideUploadContentTypeSubscriber,
            DeleteSourceFileSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        upload_call_kwargs = self.transfer_manager.upload.call_args[1]
        actual_subscribers = upload_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])


class TestDownloadRequestSubmitter(BaseTransferRequestSubmitterTest):
    def setUp(self):
        super(TestDownloadRequestSubmitter, self).setUp()
        self.transfer_request_submitter = DownloadRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )

    def assert_no_downloads_happened(self):
        self.assertEqual(len(self.transfer_manager.download.call_args_list), 0)

    def create_file_info(self, key, associated_response_data=None):
        kwargs = {
            'src': self.bucket + '/' + key,
            'src_type': 's3',
            'dest': self.filename,
            'dest_type': 'local',
            'operation_name': 'download',
            'compare_key': key,
        }
        if associated_response_data is not None:
            kwargs['associated_response_data'] = associated_response_data
        return FileInfo(**kwargs)

    def test_can_submit(self):
        fileinfo = FileInfo(
            src=self.bucket + '/' + self.key,
            dest=self.filename,
            operation_name='download',
        )
        self.assertTrue(self.transfer_request_submitter.can_submit(fileinfo))
        fileinfo.operation_name = 'foo'
        self.assertFalse(self.transfer_request_submitter.can_submit(fileinfo))

    def test_submit(self):
        fileinfo = self.create_file_info(self.key)
        future = self.transfer_request_submitter.submit(fileinfo)

        self.assertIs(self.transfer_manager.download.return_value, future)
        download_call_kwargs = self.transfer_manager.download.call_args[1]
        self.assertEqual(download_call_kwargs['fileobj'], self.filename)
        self.assertEqual(download_call_kwargs['bucket'], self.bucket)
        self.assertEqual(download_call_kwargs['key'], self.key)
        self.assertEqual(download_call_kwargs['extra_args'], {})

        # Make sure the subscriber applied are of the correct type and order
        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            DirectoryCreatorSubscriber,
            ProvideLastModifiedTimeSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = download_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_submit_with_extra_args(self):
        fileinfo = self.create_file_info(self.key)
        self.cli_params['sse_c'] = 'AES256'
        self.cli_params['sse_c_key'] = 'mykey'
        self.transfer_request_submitter.submit(fileinfo)

        # Set some extra argument like sse_c to make sure cli
        # params get mapped to request parameters.
        download_call_kwargs = self.transfer_manager.download.call_args[1]
        self.assertEqual(
            download_call_kwargs['extra_args'],
            {'SSECustomerAlgorithm': 'AES256', 'SSECustomerKey': 'mykey'},
        )

    def test_warn_glacier_for_incompatible(self):
        fileinfo = FileInfo(
            src=self.bucket + '/' + self.key,
            dest=self.filename,
            operation_name='download',
            associated_response_data={
                'StorageClass': 'GLACIER',
            },
        )
        future = self.transfer_request_submitter.submit(fileinfo)

        # A warning should have been submitted because it is a non-restored
        # glacier object.
        warning_result = self.result_queue.get()
        self.assertIsInstance(warning_result, WarningResult)
        self.assertIn(
            'Unable to perform download operations on GLACIER objects',
            warning_result.message,
        )

        # The transfer should have been skipped.
        self.assertIsNone(future)
        self.assert_no_downloads_happened()

    def test_not_warn_glacier_for_compatible(self):
        fileinfo = self.create_file_info(
            self.key,
            associated_response_data={
                'StorageClass': 'GLACIER',
                'Restore': 'ongoing-request="false"',
            },
        )
        future = self.transfer_request_submitter.submit(fileinfo)

        # A warning should have not been submitted because it is a restored
        # glacier object.
        self.assertTrue(self.result_queue.empty())

        # And the transfer should not have been skipped.
        self.assertIs(self.transfer_manager.download.return_value, future)
        self.assertEqual(len(self.transfer_manager.download.call_args_list), 1)

    def test_warn_glacier_force_glacier(self):
        self.cli_params['force_glacier_transfer'] = True
        fileinfo = self.create_file_info(
            self.key,
            associated_response_data={
                'StorageClass': 'GLACIER',
            },
        )
        future = self.transfer_request_submitter.submit(fileinfo)

        # A warning should have not been submitted because it is glacier
        # transfers were forced.
        self.assertTrue(self.result_queue.empty())
        self.assertIs(self.transfer_manager.download.return_value, future)
        self.assertEqual(len(self.transfer_manager.download.call_args_list), 1)

    def test_warn_glacier_ignore_glacier_warnings(self):
        self.cli_params['ignore_glacier_warnings'] = True
        fileinfo = FileInfo(
            src=self.bucket + '/' + self.key,
            dest=self.filename,
            operation_name='download',
            associated_response_data={
                'StorageClass': 'GLACIER',
            },
        )
        future = self.transfer_request_submitter.submit(fileinfo)

        # A warning should have not been submitted because it was specified
        # to ignore glacier warnings.
        self.assertTrue(self.result_queue.empty())
        # But the transfer still should have been skipped.
        self.assertIsNone(future)
        self.assert_no_downloads_happened()

    def test_warn_and_ignore_on_parent_dir_reference(self):
        fileinfo = self.create_file_info('../foo.txt')
        future = self.transfer_request_submitter.submit(fileinfo)
        warning_result = self.result_queue.get()
        self.assertIsInstance(warning_result, WarningResult)
        self.assert_no_downloads_happened()

    def test_warn_and_ignore_with_leading_chars(self):
        fileinfo = self.create_file_info('a/../../foo.txt')
        future = self.transfer_request_submitter.submit(fileinfo)
        warning_result = self.result_queue.get()
        self.assertIsInstance(warning_result, WarningResult)
        self.assert_no_downloads_happened()

    def test_allow_double_dots_that_dont_escape_cwd(self):
        self.cli_params['dryrun'] = True
        # This is fine because it's 'foo.txt'.
        fileinfo = self.create_file_info('a/../foo.txt')
        future = self.transfer_request_submitter.submit(fileinfo)
        self.assertIsInstance(self.result_queue.get(), DryRunResult)

    def test_dry_run(self):
        self.cli_params['dryrun'] = True
        self.transfer_request_submitter = DownloadRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )
        fileinfo = self.create_file_info(self.key)
        self.transfer_request_submitter.submit(fileinfo)

        result = self.result_queue.get()
        self.assertIsInstance(result, DryRunResult)
        self.assertEqual(result.transfer_type, 'download')
        self.assertTrue(result.dest.endswith(self.filename))
        self.assertEqual(result.src, 's3://' + self.bucket + '/' + self.key)

    def test_submit_move_adds_delete_source_subscriber(self):
        fileinfo = self.create_file_info(self.key)
        self.cli_params['guess_mime_type'] = True  # Default settings
        self.cli_params['is_move'] = True
        self.transfer_request_submitter.submit(fileinfo)
        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            DirectoryCreatorSubscriber,
            ProvideLastModifiedTimeSubscriber,
            DeleteSourceObjectSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        download_call_kwargs = self.transfer_manager.download.call_args[1]
        actual_subscribers = download_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_skip_download_when_no_overwrite_and_file_exists(self):
        self.cli_params['no_overwrite'] = True
        fileinfo = self.create_file_info(self.key)
        with mock.patch('os.path.exists', return_value=True):
            future = self.transfer_request_submitter.submit(fileinfo)

        # Result Queue should be empty because it was specified to ignore no-overwrite warnings.
        self.assertTrue(self.result_queue.empty())
        # The transfer should be skipped, so future should be None
        self.assertIsNone(future)
        self.assert_no_downloads_happened()

    def test_proceed_download_when_no_overwrite_and_file_not_exists(self):
        self.cli_params['no_overwrite'] = True
        fileinfo = self.create_file_info(self.key)
        with mock.patch('os.path.exists', return_value=False):
            future = self.transfer_request_submitter.submit(fileinfo)
        # The transfer should proceed, so future should be the transfer manager's return value
        self.assertIs(self.transfer_manager.download.return_value, future)
        # And download should have happened
        self.assertEqual(len(self.transfer_manager.download.call_args_list), 1)

    def test_warn_if_file_exists_without_no_overwrite_flag(self):
        self.cli_params['no_overwrite'] = False
        fileinfo = self.create_file_info(self.key)
        with mock.patch('os.path.exists', return_value=True):
            future = self.transfer_request_submitter.submit(fileinfo)
        # The transfer should proceed, so future should be the transfer manager's return value
        self.assertIs(self.transfer_manager.download.return_value, future)
        # And download should have happened
        self.assertEqual(len(self.transfer_manager.download.call_args_list), 1)

    def test_submit_with_full_object_checksum(self):
        fileinfo = self.create_file_info(
            self.key,
            associated_response_data={
                'ChecksumType': 'FULL_OBJECT',
                'ChecksumCRC32': 'abc123==',
            },
        )
        self.transfer_request_submitter.submit(fileinfo)
        download_call_kwargs = self.transfer_manager.download.call_args[1]
        actual_subscribers = download_call_kwargs['subscribers']
        subscriber_types = [type(s) for s in actual_subscribers]
        self.assertIn(ProvideFullObjectChecksumSubscriber, subscriber_types)

    def test_submit_without_full_object_checksum(self):
        fileinfo = self.create_file_info(self.key)
        self.transfer_request_submitter.submit(fileinfo)
        download_call_kwargs = self.transfer_manager.download.call_args[1]
        actual_subscribers = download_call_kwargs['subscribers']
        subscriber_types = [type(s) for s in actual_subscribers]
        self.assertNotIn(ProvideFullObjectChecksumSubscriber, subscriber_types)

    def test_submit_with_composite_checksum_does_not_add_subscriber(self):
        fileinfo = self.create_file_info(
            self.key,
            associated_response_data={
                'ChecksumType': 'COMPOSITE',
                'ChecksumCRC32': 'abc123==-5',
            },
        )
        self.transfer_request_submitter.submit(fileinfo)
        download_call_kwargs = self.transfer_manager.download.call_args[1]
        actual_subscribers = download_call_kwargs['subscribers']
        subscriber_types = [type(s) for s in actual_subscribers]
        self.assertNotIn(ProvideFullObjectChecksumSubscriber, subscriber_types)


class TestCopyRequestSubmitter(BaseTransferRequestSubmitterTest):
    def setUp(self):
        super(TestCopyRequestSubmitter, self).setUp()
        self.source_bucket = 'mysourcebucket'
        self.source_key = 'mysourcekey'
        self.transfer_request_submitter = CopyRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )

    def test_can_submit(self):
        fileinfo = FileInfo(
            src=self.source_bucket + '/' + self.source_key,
            dest=self.bucket + '/' + self.key,
            operation_name='copy',
        )
        self.assertTrue(self.transfer_request_submitter.can_submit(fileinfo))
        fileinfo.operation_name = 'foo'
        self.assertFalse(self.transfer_request_submitter.can_submit(fileinfo))

    def test_submit(self):
        fileinfo = FileInfo(
            src=self.source_bucket + '/' + self.source_key,
            dest=self.bucket + '/' + self.key,
        )
        self.cli_params['guess_mime_type'] = True  # Default settings
        future = self.transfer_request_submitter.submit(fileinfo)
        self.assertIs(self.transfer_manager.copy.return_value, future)
        copy_call_kwargs = self.transfer_manager.copy.call_args[1]
        self.assertEqual(
            copy_call_kwargs['copy_source'],
            {'Bucket': self.source_bucket, 'Key': self.source_key},
        )
        self.assertEqual(copy_call_kwargs['bucket'], self.bucket)
        self.assertEqual(copy_call_kwargs['key'], self.key)
        self.assertEqual(copy_call_kwargs['extra_args'], {})

        # Make sure the subscriber applied are of the correct type and order
        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            SetMetadataDirectivePropsSubscriber,
            SetTagsSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = copy_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_submit_with_extra_args(self):
        fileinfo = FileInfo(
            src=self.source_bucket + '/' + self.source_key,
            dest=self.bucket + '/' + self.key,
        )
        # Set some extra argument like storage_class to make sure cli
        # params get mapped to request parameters.
        self.cli_params['storage_class'] = 'STANDARD_IA'
        self.transfer_request_submitter.submit(fileinfo)

        copy_call_kwargs = self.transfer_manager.copy.call_args[1]
        self.assertEqual(
            copy_call_kwargs['extra_args'], {'StorageClass': 'STANDARD_IA'}
        )

    def test_submit_when_content_type_specified(self):
        fileinfo = FileInfo(
            src=self.source_bucket + '/' + self.source_key,
            dest=self.bucket + '/' + self.key,
        )
        self.cli_params['content_type'] = 'text/plain'
        self.transfer_request_submitter.submit(fileinfo)

        copy_call_kwargs = self.transfer_manager.copy.call_args[1]
        self.assertEqual(
            copy_call_kwargs['extra_args'], {'ContentType': 'text/plain'}
        )
        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            SetMetadataDirectivePropsSubscriber,
            SetTagsSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = copy_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_metadata_directive_excludes_copy_props_subscribers(self):
        fileinfo = FileInfo(
            src=self.source_bucket + '/' + self.source_key,
            dest=self.bucket + '/' + self.key,
        )
        self.cli_params['copy_props'] = 'default'
        self.cli_params['metadata_directive'] = 'REPLACE'
        self.transfer_request_submitter.submit(fileinfo)

        copy_call_kwargs = self.transfer_manager.copy.call_args[1]
        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = copy_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_warn_glacier_for_incompatible(self):
        fileinfo = FileInfo(
            src=self.source_bucket + '/' + self.source_key,
            dest=self.bucket + '/' + self.key,
            operation_name='copy',
            associated_response_data={
                'StorageClass': 'GLACIER',
            },
        )
        future = self.transfer_request_submitter.submit(fileinfo)

        # A warning should have been submitted because it is a non-restored
        # glacier object.
        warning_result = self.result_queue.get()
        self.assertIsInstance(warning_result, WarningResult)
        self.assertIn(
            'Unable to perform copy operations on GLACIER objects',
            warning_result.message,
        )

        # The transfer request should have never been sent therefore return
        # no future.
        self.assertIsNone(future)
        # The transfer should have been skipped.
        self.assertEqual(len(self.transfer_manager.copy.call_args_list), 0)

    def test_not_warn_glacier_for_compatible(self):
        fileinfo = FileInfo(
            src=self.source_bucket + '/' + self.source_key,
            dest=self.bucket + '/' + self.key,
            operation_name='copy',
            associated_response_data={
                'StorageClass': 'GLACIER',
                'Restore': 'ongoing-request="false"',
            },
        )
        future = self.transfer_request_submitter.submit(fileinfo)
        self.assertIs(self.transfer_manager.copy.return_value, future)

        # A warning should have not been submitted because it is a restored
        # glacier object.
        self.assertTrue(self.result_queue.empty())

        # And the transfer should not have been skipped.
        self.assertEqual(len(self.transfer_manager.copy.call_args_list), 1)

    def test_warn_glacier_force_glacier(self):
        self.cli_params['force_glacier_transfer'] = True
        fileinfo = FileInfo(
            src=self.source_bucket + '/' + self.source_key,
            dest=self.bucket + '/' + self.key,
            operation_name='copy',
            associated_response_data={
                'StorageClass': 'GLACIER',
            },
        )
        future = self.transfer_request_submitter.submit(fileinfo)
        self.assertIs(self.transfer_manager.copy.return_value, future)

        # A warning should have not been submitted because it is glacier
        # transfers were forced.
        self.assertTrue(self.result_queue.empty())
        self.assertEqual(len(self.transfer_manager.copy.call_args_list), 1)

    def test_warn_glacier_ignore_glacier_warnings(self):
        self.cli_params['ignore_glacier_warnings'] = True
        fileinfo = FileInfo(
            src=self.source_bucket + '/' + self.source_key,
            dest=self.bucket + '/' + self.key,
            operation_name='copy',
            associated_response_data={
                'StorageClass': 'GLACIER',
            },
        )
        future = self.transfer_request_submitter.submit(fileinfo)

        # The transfer request should have never been sent therefore return
        # no future.
        self.assertIsNone(future)
        # A warning should have not been submitted because it was specified
        # to ignore glacier warnings.
        self.assertTrue(self.result_queue.empty())
        # But the transfer still should have been skipped.
        self.assertEqual(len(self.transfer_manager.copy.call_args_list), 0)

    def test_dry_run(self):
        self.cli_params['dryrun'] = True
        self.transfer_request_submitter = CopyRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )
        fileinfo = FileInfo(
            src=self.source_bucket + '/' + self.source_key,
            src_type='s3',
            dest=self.bucket + '/' + self.key,
            dest_type='s3',
            operation_name='copy',
        )
        self.transfer_request_submitter.submit(fileinfo)

        result = self.result_queue.get()
        self.assertIsInstance(result, DryRunResult)
        self.assertEqual(result.transfer_type, 'copy')
        source = 's3://' + self.source_bucket + '/' + self.source_key
        self.assertEqual(result.src, source)
        self.assertEqual(result.dest, 's3://' + self.bucket + '/' + self.key)

    def test_submit_move_adds_delete_source_subscriber(self):
        fileinfo = FileInfo(
            dest=self.source_bucket + '/' + self.source_key,
            src=self.bucket + '/' + self.key,
        )
        self.cli_params['guess_mime_type'] = True  # Default settings
        self.cli_params['is_move'] = True
        self.transfer_request_submitter.submit(fileinfo)
        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            SetMetadataDirectivePropsSubscriber,
            SetTagsSubscriber,
            DeleteSourceObjectSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        copy_call_kwargs = self.transfer_manager.copy.call_args[1]
        actual_subscribers = copy_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_file_exists_without_no_overwrite(self):
        self.cli_params['no_overwrite'] = False
        fileinfo = FileInfo(
            src=self.source_bucket + "/" + self.source_key,
            dest=self.bucket + "/" + self.key,
            operation_name='copy',
            size=100,
            source_client=mock.Mock(),
        )
        future = self.transfer_request_submitter.submit(fileinfo)
        # The transfer should proceed, so future should be the transfer manager's return value
        self.assertIs(self.transfer_manager.copy.return_value, future)
        self.assertEqual(len(self.transfer_manager.copy.call_args_list), 1)
        # Head should not be called when no_overwrite is false
        fileinfo.source_client.head_object.assert_not_called()


class TestUploadStreamRequestSubmitter(BaseTransferRequestSubmitterTest):
    def setUp(self):
        super(TestUploadStreamRequestSubmitter, self).setUp()
        self.filename = '-'
        self.cli_params['is_stream'] = True
        self.transfer_request_submitter = UploadStreamRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )

    def test_can_submit(self):
        fileinfo = FileInfo(
            src=self.filename,
            dest=self.bucket + '/' + self.key,
            operation_name='upload',
        )
        self.assertTrue(self.transfer_request_submitter.can_submit(fileinfo))
        self.cli_params['is_stream'] = False
        self.assertFalse(self.transfer_request_submitter.can_submit(fileinfo))

    def test_submit(self):
        fileinfo = FileInfo(
            src=self.filename, dest=self.bucket + '/' + self.key
        )
        future = self.transfer_request_submitter.submit(fileinfo)
        self.assertIs(self.transfer_manager.upload.return_value, future)

        upload_call_kwargs = self.transfer_manager.upload.call_args[1]
        self.assertIsInstance(upload_call_kwargs['fileobj'], NonSeekableStream)
        self.assertEqual(upload_call_kwargs['bucket'], self.bucket)
        self.assertEqual(upload_call_kwargs['key'], self.key)
        self.assertEqual(upload_call_kwargs['extra_args'], {})

        ref_subscribers = [
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = upload_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_submit_with_expected_size_provided(self):
        provided_size = 100
        self.cli_params['expected_size'] = provided_size
        fileinfo = FileInfo(
            src=self.filename, dest=self.bucket + '/' + self.key
        )
        self.transfer_request_submitter.submit(fileinfo)
        upload_call_kwargs = self.transfer_manager.upload.call_args[1]

        ref_subscribers = [
            ProvideSizeSubscriber,
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = upload_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])
        # The ProvideSizeSubscriber should be providing the correct size
        self.assertEqual(actual_subscribers[0].size, provided_size)

    def test_dry_run(self):
        self.cli_params['dryrun'] = True
        self.transfer_request_submitter = UploadStreamRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )
        fileinfo = FileInfo(
            src=self.filename,
            src_type='local',
            operation_name='upload',
            dest=self.bucket + '/' + self.key,
            dest_type='s3',
        )
        self.transfer_request_submitter.submit(fileinfo)

        result = self.result_queue.get()
        self.assertIsInstance(result, DryRunResult)
        self.assertEqual(result.transfer_type, 'upload')
        self.assertEqual(result.dest, 's3://' + self.bucket + '/' + self.key)
        self.assertEqual(result.src, '-')


class TestDownloadStreamRequestSubmitter(BaseTransferRequestSubmitterTest):
    def setUp(self):
        super(TestDownloadStreamRequestSubmitter, self).setUp()
        self.filename = '-'
        self.cli_params['is_stream'] = True
        self.transfer_request_submitter = DownloadStreamRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )

    def test_can_submit(self):
        fileinfo = FileInfo(
            src=self.bucket + '/' + self.key,
            dest=self.filename,
            operation_name='download',
        )
        self.assertTrue(self.transfer_request_submitter.can_submit(fileinfo))
        self.cli_params['is_stream'] = False
        self.assertFalse(self.transfer_request_submitter.can_submit(fileinfo))

    def test_submit(self):
        fileinfo = FileInfo(
            src=self.bucket + '/' + self.key,
            dest=self.filename,
            compare_key=self.key,
        )
        future = self.transfer_request_submitter.submit(fileinfo)
        self.assertIs(self.transfer_manager.download.return_value, future)

        download_call_kwargs = self.transfer_manager.download.call_args[1]
        self.assertIsInstance(
            download_call_kwargs['fileobj'], StdoutBytesWriter
        )
        self.assertEqual(download_call_kwargs['bucket'], self.bucket)
        self.assertEqual(download_call_kwargs['key'], self.key)
        self.assertEqual(download_call_kwargs['extra_args'], {})

        ref_subscribers = [
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = download_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_dry_run(self):
        self.cli_params['dryrun'] = True
        self.transfer_request_submitter = DownloadStreamRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )
        fileinfo = FileInfo(
            dest=self.filename,
            dest_type='local',
            operation_name='download',
            src=self.bucket + '/' + self.key,
            src_type='s3',
            compare_key=self.key,
        )
        self.transfer_request_submitter.submit(fileinfo)

        result = self.result_queue.get()
        self.assertIsInstance(result, DryRunResult)
        self.assertEqual(result.transfer_type, 'download')
        self.assertEqual(result.src, 's3://' + self.bucket + '/' + self.key)
        self.assertEqual(result.dest, '-')


class TestDeleteRequestSubmitter(BaseTransferRequestSubmitterTest):
    def setUp(self):
        super(TestDeleteRequestSubmitter, self).setUp()
        self.transfer_request_submitter = DeleteRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )

    def test_can_submit(self):
        fileinfo = FileInfo(
            src=self.bucket + '/' + self.key,
            dest=None,
            operation_name='delete',
            src_type='s3',
        )
        self.assertTrue(self.transfer_request_submitter.can_submit(fileinfo))
        fileinfo.operation_name = 'foo'
        self.assertFalse(self.transfer_request_submitter.can_submit(fileinfo))

    def test_cannot_submit_local_deletes(self):
        fileinfo = FileInfo(
            src=self.bucket + '/' + self.key,
            dest=None,
            operation_name='delete',
            src_type='local',
        )
        self.assertFalse(self.transfer_request_submitter.can_submit(fileinfo))

    def test_submit(self):
        fileinfo = FileInfo(
            src=self.bucket + '/' + self.key,
            dest=None,
            operation_name='delete',
        )
        future = self.transfer_request_submitter.submit(fileinfo)
        self.assertIs(self.transfer_manager.delete.return_value, future)

        delete_call_kwargs = self.transfer_manager.delete.call_args[1]
        self.assertEqual(delete_call_kwargs['bucket'], self.bucket)
        self.assertEqual(delete_call_kwargs['key'], self.key)
        self.assertEqual(delete_call_kwargs['extra_args'], {})

        ref_subscribers = [
            ProvideETagSubscriber,
            QueuedResultSubscriber,
            ProgressResultSubscriber,
            DoneResultSubscriber,
        ]
        actual_subscribers = delete_call_kwargs['subscribers']
        self.assertEqual(len(ref_subscribers), len(actual_subscribers))
        for i, actual_subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(actual_subscriber, ref_subscribers[i])

    def test_dry_run(self):
        self.cli_params['dryrun'] = True
        self.transfer_request_submitter = DeleteRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )
        fileinfo = FileInfo(
            src=self.bucket + '/' + self.key,
            src_type='s3',
            dest=self.bucket + '/' + self.key,
            dest_type='s3',
            operation_name='delete',
        )
        self.transfer_request_submitter.submit(fileinfo)

        result = self.result_queue.get()
        self.assertIsInstance(result, DryRunResult)
        self.assertEqual(result.transfer_type, 'delete')
        self.assertEqual(result.src, 's3://' + self.bucket + '/' + self.key)
        self.assertIsNone(result.dest)


class TestLocalDeleteRequestSubmitter(BaseTransferRequestSubmitterTest):
    def setUp(self):
        super(TestLocalDeleteRequestSubmitter, self).setUp()
        self.transfer_request_submitter = LocalDeleteRequestSubmitter(
            self.transfer_manager, self.result_queue, self.cli_params
        )
        self.file_creator = FileCreator()

    def tearDown(self):
        super(TestLocalDeleteRequestSubmitter, self).tearDown()
        self.file_creator.remove_all()

    def test_can_submit(self):
        fileinfo = FileInfo(
            src=self.filename,
            dest=None,
            operation_name='delete',
            src_type='local',
        )
        self.assertTrue(self.transfer_request_submitter.can_submit(fileinfo))
        fileinfo.operation_name = 'foo'
        self.assertFalse(self.transfer_request_submitter.can_submit(fileinfo))

    def test_cannot_submit_remote_deletes(self):
        fileinfo = FileInfo(
            src=self.filename,
            dest=None,
            operation_name='delete',
            src_type='s3',
        )
        self.assertFalse(self.transfer_request_submitter.can_submit(fileinfo))

    def test_submit(self):
        full_filename = self.file_creator.create_file(self.filename, 'content')
        fileinfo = FileInfo(
            src=full_filename,
            dest=None,
            operation_name='delete',
            src_type='local',
        )
        rval = self.transfer_request_submitter.submit(fileinfo)
        self.assertTrue(rval)

        queued_result = self.result_queue.get()
        self.assertIsInstance(queued_result, QueuedResult)
        self.assertEqual(queued_result.transfer_type, 'delete')
        self.assertTrue(queued_result.src.endswith(self.filename))
        self.assertIsNone(queued_result.dest)
        self.assertEqual(queued_result.total_transfer_size, 0)

        failure_result = self.result_queue.get()
        self.assertIsInstance(failure_result, SuccessResult)
        self.assertEqual(failure_result.transfer_type, 'delete')
        self.assertTrue(failure_result.src.endswith(self.filename))
        self.assertIsNone(failure_result.dest)

        self.assertFalse(os.path.exists(full_filename))

    def test_submit_with_exception(self):
        fileinfo = FileInfo(
            src=self.filename,
            dest=None,
            operation_name='delete',
            src_type='local',
        )
        # The file was never created so it should trigger an exception
        # when it is attempted to be deleted in the submitter.
        rval = self.transfer_request_submitter.submit(fileinfo)
        self.assertTrue(rval)

        queued_result = self.result_queue.get()
        self.assertIsInstance(queued_result, QueuedResult)
        self.assertEqual(queued_result.transfer_type, 'delete')
        self.assertTrue(queued_result.src.endswith(self.filename))
        self.assertIsNone(queued_result.dest)
        self.assertEqual(queued_result.total_transfer_size, 0)

        failure_result = self.result_queue.get()
        self.assertIsInstance(failure_result, FailureResult)
        self.assertEqual(failure_result.transfer_type, 'delete')
        self.assertTrue(failure_result.src.endswith(self.filename))
        self.assertIsNone(failure_result.dest)

    def test_dry_run(self):
        self.cli_params['dryrun'] = True
        fileinfo = FileInfo(
            src=self.filename,
            src_type='local',
            dest=self.filename,
            dest_type='local',
            operation_name='delete',
        )
        self.transfer_request_submitter.submit(fileinfo)

        result = self.result_queue.get()
        self.assertIsInstance(result, DryRunResult)
        self.assertEqual(result.transfer_type, 'delete')
        self.assertTrue(result.src.endswith(self.filename))
        self.assertIsNone(result.dest)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.arguments import CLIArgument
from awscli.customizations import utils
from awscli.customizations.flatten import FlattenArguments, FlattenedArgument
from awscli.testutils import mock, unittest


def _hydrate(params, container, cli_type, key, value):
    """
    An example to hydrate a complex structure with custom value logic. In this
    case we create a nested structure and divide the value by 100.
    """
    params['bag'] = {'ArgumentBaz': {'SomeValueAbc': value / 100.0}}


FLATTEN_CONFIG = {
    'command-name': {
        'original-argument': {
            "keep": False,
            "flatten": {
                "ArgumentFoo": {"name": "foo"},
                "ArgumentBar": {
                    "name": "bar",
                    "help_text": "Some help text",
                    "required": True,
                    "hydrate_value": lambda x: x.upper(),
                },
            },
        },
        'another-original-argument': {
            "keep": True,
            "flatten": {
                "ArgumentBaz.SomeValue": {"name": "baz", "hydrate": _hydrate}
            },
        },
    }
}


class TestFlattenedArgument(unittest.TestCase):
    def test_basic_argument(self):
        container = mock.Mock()
        container.argument_model.name = 'bag'
        kwargs = {'container': container, 'prop': 'ArgumentFoo'}
        kwargs['container'].py_name = 'bag'
        kwargs.update(
            FLATTEN_CONFIG['command-name']['original-argument']['flatten'][
                'ArgumentFoo'
            ]
        )
        arg = FlattenedArgument(**kwargs)

        self.assertEqual('foo', arg.name)
        self.assertEqual('', arg.documentation)
        self.assertEqual(False, arg.required)

        params = {}
        arg.add_to_params(params, 'value')
        self.assertEqual('value', params['bag']['ArgumentFoo'])

    def test_hydrate_value_argument(self):
        container = mock.Mock()
        container.argument_model.name = 'bag'
        kwargs = {'container': container, 'prop': 'ArgumentBar'}
        kwargs['container'].py_name = 'bag'
        kwargs['container'].cli_type_name = 'list'
        kwargs.update(
            FLATTEN_CONFIG['command-name']['original-argument']['flatten'][
                'ArgumentBar'
            ]
        )
        arg = FlattenedArgument(**kwargs)

        self.assertEqual('bar', arg.name)
        self.assertEqual('Some help text', arg.documentation)
        self.assertEqual(True, arg.required)

        params = {}
        arg.add_to_params(params, 'value')
        self.assertEqual('VALUE', params['bag'][0]['ArgumentBar'])

    def test_hydrate_function_argument(self):
        container = mock.Mock()
        container.argument_model.name = 'bag'
        kwargs = {'container': container, 'prop': 'ArgumentBaz:SomeValue'}
        kwargs['container'].py_name = 'bag'
        kwargs.update(
            FLATTEN_CONFIG['command-name']['another-original-argument'][
                'flatten'
            ]['ArgumentBaz.SomeValue']
        )
        arg = FlattenedArgument(**kwargs)

        self.assertEqual('baz', arg.name)
        self.assertEqual('', arg.documentation)
        self.assertEqual(False, arg.required)

        params = {}
        arg.add_to_params(params, 1020)
        self.assertEqual(10.2, params['bag']['ArgumentBaz']['SomeValueAbc'])


class TestFlattenCommands(unittest.TestCase):
    def test_flatten_register(self):
        cli = mock.Mock()

        flatten = FlattenArguments('my-service', FLATTEN_CONFIG)
        flatten.register(cli)

        cli.register.assert_called_with(
            'building-argument-table.my-service.command-name',
            flatten.flatten_args,
        )

    def test_flatten_modify_args(self):
        # Mock operation, arguments, and members for a service
        command = mock.Mock()
        command.name = 'command-name'

        argument_model1 = mock.Mock()
        argument_model1.required_members = []

        member_foo = mock.Mock()
        member_foo.name = 'ArgumentFoo'
        member_foo.documentation = 'Original docs'
        member_foo.required_members = []

        member_bar = mock.Mock()
        member_bar.name = 'ArgumentBar'
        member_bar.documentation = 'More docs'
        member_bar.required_members = []

        argument_model1.members = {
            'ArgumentFoo': member_foo,
            'ArgumentBar': member_bar,
        }

        argument_model2 = mock.Mock()
        argument_model2.required_members = []

        member_baz = mock.Mock()
        member_baz.name = 'ArgumentBaz'
        member_baz.documentation = ''
        member_baz.required_members = []

        member_some_value = mock.Mock()
        member_some_value.name = 'SomeValue'
        member_some_value.documenation = ''
        member_some_value.required_members = []

        member_baz.members = {'SomeValue': member_some_value}

        argument_model2.members = {'ArgumentBaz': member_baz}

        cli_argument1 = mock.Mock(spec=CLIArgument)
        cli_argument1.argument_model = argument_model1

        cli_argument2 = mock.Mock(spec=CLIArgument)
        cli_argument2.argument_model = argument_model2

        argument_table = {
            'original-argument': cli_argument1,
            'another-original-argument': cli_argument2,
        }

        # Create the flattened argument table
        cli = mock.Mock()
        flatten = FlattenArguments('my-service', FLATTEN_CONFIG)
        flatten.flatten_args(command, argument_table)

        # Make sure new arguments and any with keep=True are there
        self.assertIn('foo', argument_table)
        self.assertIn('bar', argument_table)
        self.assertNotIn('original-argument', argument_table)
        self.assertIn('baz', argument_table)
        self.assertIn('another-original-argument', argument_table)

        # Make sure the new arguments are the class we expect
        self.assertIsInstance(argument_table['foo'], FlattenedArgument)
        self.assertIsInstance(argument_table['bar'], FlattenedArgument)
        self.assertIsInstance(argument_table['baz'], FlattenedArgument)
        self.assertNotIsInstance(
            argument_table['another-original-argument'], FlattenedArgument
        )

        # Make sure original required trait can be overridden
        self.assertEqual(False, argument_table['foo'].required)
        self.assertEqual(True, argument_table['bar'].required)

        # Make sure docs can be overriden and get the defaults
        self.assertEqual('Original docs', argument_table['foo'].documentation)
        self.assertEqual('Some help text', argument_table['bar'].documentation)



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

from awscli.compat import BytesIO
from awscli.testutils import BaseAWSCommandParamsTest, FileCreator, mock


class TestGetGameSessionLog(BaseAWSCommandParamsTest):
    prefix = 'gamelift get-game-session-log'

    def setUp(self):
        super(TestGetGameSessionLog, self).setUp()
        self.files = FileCreator()
        self.filename = os.path.join(self.files.rootdir, 'myfile')
        self.urlopen_patch = mock.patch(
            'awscli.customizations.gamelift.getlog.urlopen'
        )
        self.contents = b'My Contents'
        self.urlopen_mock = self.urlopen_patch.start()
        self.urlopen_mock.return_value = BytesIO(self.contents)

    def tearDown(self):
        super(TestGetGameSessionLog, self).tearDown()
        self.files.remove_all()
        self.urlopen_patch.stop()

    def test_get_game_session_log(self):
        cmdline = self.prefix
        cmdline += ' --game-session-id mysession'
        cmdline += ' --save-as %s' % self.filename

        self.parsed_responses = [{'PreSignedUrl': 'myurl'}]
        stdout, stderr, rc = self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(
            self.operations_called[0][0].name, 'GetGameSessionLogUrl'
        )
        self.assertEqual(
            self.operations_called[0][1], {'GameSessionId': 'mysession'}
        )

        # Ensure the contents were saved to the file
        self.assertTrue(os.path.exists(self.filename))
        with open(self.filename, 'rb') as f:
            self.assertEqual(f.read(), self.contents)

        # Ensure the output is as expected
        self.assertIn(
            'Successfully downloaded log archive for game session '
            'mysession to %s' % self.filename,
            stdout,
        )



import os

import pytest

from awscli.clidriver import create_clidriver
from awscli.testutils import FileCreator, mock


def set_up(config, env_var):
    files = FileCreator()
    environ = {
        'AWS_DATA_PATH': os.environ['AWS_DATA_PATH'],
        'AWS_DEFAULT_REGION': 'us-east-1',
        'AWS_ACCESS_KEY_ID': 'access_key',
        'AWS_SECRET_ACCESS_KEY': 'secret_key',
        'AWS_CONFIG_FILE': '',
    }
    environ = set_config_file_contents(
        environ=environ, config=config, files=files
    )
    environ = set_env_var(environ=environ, env_var=env_var)
    environ_patch = mock.patch('os.environ', environ)
    environ_patch.start()
    driver = create_clidriver()
    return driver, environ_patch, files


def set_config_file_contents(environ, files, config=None):
    config_contents = ''
    if config is not None:
        config_contents = '[default]\n' f'cli_auto_prompt = {config}\n'
    environ['AWS_CONFIG_FILE'] = files.create_file('config', config_contents)
    return environ


def set_env_var(environ, env_var=None):
    if env_var is not None:
        environ['AWS_CLI_AUTO_PROMPT'] = env_var
    return environ


@pytest.mark.parametrize(
    "config,env_var,expected_result",
    [
        ('on', 'on', 'on'),
        ('on', 'off', 'off'),
        ('on', None, 'on'),
        ('off', 'on', 'on'),
        ('off', 'off', 'off'),
        ('off', None, 'off'),
        (None, 'on', 'on'),
        (None, 'off', 'off'),
        (None, None, 'off'),
    ],
)
def test_autoprompt_config_provider(config, env_var, expected_result):
    try:
        driver, environ_patch, files = set_up(config, env_var)
        _assert_auto_prompt_configures_as_expected(driver, expected_result)
    finally:
        environ_patch.stop()
        files.remove_all()


def _assert_auto_prompt_configures_as_expected(driver, expected_result):
    actual = driver.session.get_config_variable('cli_auto_prompt')
    assert actual == expected_result



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from prompt_toolkit.application import DummyApplication
from prompt_toolkit.layout import walk

from awscli.customizations.wizard.ui.layout import (
    RunWizardDialog,
    WizardErrorBar,
)
from awscli.testutils import mock, unittest


class TestRunWizardDialog(unittest.TestCase):
    def get_title_text(self, dialog):
        container = dialog.container
        frame_label = self.get_child_containers(container, 'frame.label')[0]
        # Frame label texts are wrapped in a couple of lambdas that need to be
        # unrolled.
        return self._build_text_from_ui_control(frame_label.content.text()())

    def get_button_texts(self, dialog):
        container = dialog.container
        button_texts = []
        for button in self.get_child_containers(container, 'button'):
            button_texts.append(
                self._build_text_from_ui_control(button.content.text)
            )
        return button_texts

    def get_child_containers(self, container, with_style_cls=None):
        child_containers = []
        for child_container in walk(container):
            if with_style_cls:
                if self._matches_style_cls(child_container, with_style_cls):
                    child_containers.append(child_container)
            else:
                child_containers.append(child_container)
        return child_containers

    def _matches_style_cls(self, container, style_cls):
        if hasattr(container, 'style'):
            container_style = container.style
            if callable(container_style):
                container_style = container_style()
            return f'class:{style_cls}' in container_style
        return False

    def _build_text_from_ui_control(self, ui_control_text):
        text = ''
        for text_component in ui_control_text():
            text += text_component[1]
        return text

    def test_from_done_section_definition(self):
        done_section = {}
        dialog = RunWizardDialog.from_done_section_definition(done_section)
        self.assertIn('Run wizard?', self.get_title_text(dialog))
        self.assertEqual(
            self.get_button_texts(dialog), ['<   Yes    >', '<   Back   >']
        )

    def test_override_title_in_done_section_definition(self):
        done_section = {'description': 'Override title'}
        dialog = RunWizardDialog.from_done_section_definition(done_section)
        self.assertIn('Override title', self.get_title_text(dialog))

    def test_override_button_list_in_done_section_definition(self):
        done_section = {'options': ['yes']}
        dialog = RunWizardDialog.from_done_section_definition(done_section)
        self.assertEqual(
            self.get_button_texts(dialog),
            [
                '<   Yes    >',
            ],
        )

    def test_override_button_text_in_done_section_definition(self):
        done_section = {
            'options': [
                {'name': 'yes', 'description': 'Y'},
                {'name': 'back', 'description': 'B'},
            ]
        }
        dialog = RunWizardDialog.from_done_section_definition(done_section)
        self.assertEqual(
            self.get_button_texts(dialog),
            [
                '<    Y     >',
                '<    B     >',
            ],
        )


class TestWizardErrorBar(unittest.TestCase):
    def setUp(self):
        self.error_bar = WizardErrorBar()
        self.app = DummyApplication()
        self.app.error_bar_visible = None
        self.get_app = mock.patch(
            'awscli.customizations.wizard.ui.layout.get_app',
            return_value=self.app,
        )
        self.get_app.start()

    def tearDown(self):
        self.get_app.stop()

    def assert_error_bar_is_visible(self):
        self.assertTrue(self.error_bar.container.filter())

    def assert_error_bar_is_not_visible(self):
        self.assertFalse(self.error_bar.container.filter())

    def get_error_bar_message(self):
        return self.get_error_bar_buffer().text

    def get_error_bar_buffer(self):
        for child in walk(self.error_bar.container, skip_hidden=True):
            if (
                hasattr(child, 'content')
                and hasattr(child.content, 'buffer')
                and child.content.buffer.name == 'error_bar'
            ):
                return child.content.buffer
        return None

    def test_visible_when_exception(self):
        self.error_bar.display_error(Exception('Error message'))
        self.assert_error_bar_is_visible()
        self.assertIn('Error message', self.get_error_bar_message())

    def test_not_visible_by_default(self):
        self.assert_error_bar_is_not_visible()

    def test_can_clear_error_bar(self):
        self.error_bar.display_error(Exception('Error message'))
        self.assert_error_bar_is_visible()
        self.error_bar.clear()
        self.assert_error_bar_is_not_visible()

    def test_can_override_existing_error(self):
        self.error_bar.display_error(Exception('Original'))
        self.error_bar.display_error(Exception('Override'))
        error_bar_message = self.get_error_bar_message()
        self.assertNotIn('Original', error_bar_message)
        self.assertIn('Override', error_bar_message)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.model import OperationModel, ServiceModel

from awscli.customizations.dynamodb.transform import ParameterTransformer
from awscli.testutils import unittest


class BaseTransformationTest(unittest.TestCase):
    def setUp(self):
        self.target_shape = 'MyShape'
        self.original_value = 'orginal'
        self.transformed_value = 'transformed'
        self.transformer = ParameterTransformer()
        self.json_model = {}
        self.nested_json_model = {}
        self.setup_models()
        self.build_models()

    def setup_models(self):
        self.json_model = {
            'operations': {
                'SampleOperation': {
                    'name': 'SampleOperation',
                    'input': {'shape': 'SampleOperationInputOutput'},
                    'output': {'shape': 'SampleOperationInputOutput'},
                }
            },
            'shapes': {
                'SampleOperationInputOutput': {
                    'type': 'structure',
                    'members': {},
                },
                'String': {'type': 'string'},
            },
        }

    def build_models(self):
        self.service_model = ServiceModel(self.json_model)
        self.operation_model = OperationModel(
            self.json_model['operations']['SampleOperation'],
            self.service_model,
        )

    def add_input_shape(self, shape):
        self.add_shape(shape)
        params_shape = self.json_model['shapes']['SampleOperationInputOutput']
        shape_name = list(shape.keys())[0]
        params_shape['members'][shape_name] = {'shape': shape_name}

    def add_shape(self, shape):
        shape_name = list(shape.keys())[0]
        self.json_model['shapes'][shape_name] = shape[shape_name]


class TestInputOutputTransformer(BaseTransformationTest):
    def setUp(self):
        super(TestInputOutputTransformer, self).setUp()
        self.transformation = lambda params: self.transformed_value
        self.add_shape({self.target_shape: {'type': 'string'}})

    def test_transform_structure(self):
        input_params = {
            'Structure': {
                'TransformMe': self.original_value,
                'LeaveAlone': self.original_value,
            }
        }
        input_shape = {
            'Structure': {
                'type': 'structure',
                'members': {
                    'TransformMe': {'shape': self.target_shape},
                    'LeaveAlone': {'shape': 'String'},
                },
            }
        }

        self.add_input_shape(input_shape)
        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        self.assertEqual(
            input_params,
            {
                'Structure': {
                    'TransformMe': self.transformed_value,
                    'LeaveAlone': self.original_value,
                }
            },
        )

    def test_transform_map(self):
        input_params = {
            'TransformMe': {'foo': self.original_value},
            'LeaveAlone': {'foo': self.original_value},
        }

        targeted_input_shape = {
            'TransformMe': {
                'type': 'map',
                'key': {'shape': 'String'},
                'value': {'shape': self.target_shape},
            }
        }

        untargeted_input_shape = {
            'LeaveAlone': {
                'type': 'map',
                'key': {'shape': 'String'},
                'value': {'shape': 'String'},
            }
        }

        self.add_input_shape(targeted_input_shape)
        self.add_input_shape(untargeted_input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        self.assertEqual(
            input_params,
            {
                'TransformMe': {'foo': self.transformed_value},
                'LeaveAlone': {'foo': self.original_value},
            },
        )

    def test_transform_list(self):
        input_params = {
            'TransformMe': [self.original_value, self.original_value],
            'LeaveAlone': [self.original_value, self.original_value],
        }

        targeted_input_shape = {
            'TransformMe': {
                'type': 'list',
                'member': {'shape': self.target_shape},
            }
        }

        untargeted_input_shape = {
            'LeaveAlone': {'type': 'list', 'member': {'shape': 'String'}}
        }

        self.add_input_shape(targeted_input_shape)
        self.add_input_shape(untargeted_input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        self.assertEqual(
            input_params,
            {
                'TransformMe': [
                    self.transformed_value,
                    self.transformed_value,
                ],
                'LeaveAlone': [self.original_value, self.original_value],
            },
        )

    def test_transform_nested_structure(self):
        input_params = {
            'WrapperStructure': {
                'Structure': {
                    'TransformMe': self.original_value,
                    'LeaveAlone': self.original_value,
                }
            }
        }

        structure_shape = {
            'Structure': {
                'type': 'structure',
                'members': {
                    'TransformMe': {'shape': self.target_shape},
                    'LeaveAlone': {'shape': 'String'},
                },
            }
        }

        input_shape = {
            'WrapperStructure': {
                'type': 'structure',
                'members': {'Structure': {'shape': 'Structure'}},
            }
        }
        self.add_shape(structure_shape)
        self.add_input_shape(input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        self.assertEqual(
            input_params,
            {
                'WrapperStructure': {
                    'Structure': {
                        'TransformMe': self.transformed_value,
                        'LeaveAlone': self.original_value,
                    }
                }
            },
        )

    def test_transform_nested_map(self):
        input_params = {
            'TargetedWrapperMap': {'foo': {'bar': self.original_value}},
            'UntargetedWrapperMap': {'foo': {'bar': self.original_value}},
        }

        targeted_map_shape = {
            'TransformMeMap': {
                'type': 'map',
                'key': {'shape': 'String'},
                'value': {'shape': self.target_shape},
            }
        }

        targeted_wrapper_shape = {
            'TargetedWrapperMap': {
                'type': 'map',
                'key': {'shape': 'Name'},
                'value': {'shape': 'TransformMeMap'},
            }
        }

        self.add_shape(targeted_map_shape)
        self.add_input_shape(targeted_wrapper_shape)

        untargeted_map_shape = {
            'LeaveAloneMap': {
                'type': 'map',
                'key': {'shape': 'String'},
                'value': {'shape': 'String'},
            }
        }

        untargeted_wrapper_shape = {
            'UntargetedWrapperMap': {
                'type': 'map',
                'key': {'shape': 'Name'},
                'value': {'shape': 'LeaveAloneMap'},
            }
        }

        self.add_shape(untargeted_map_shape)
        self.add_input_shape(untargeted_wrapper_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        self.assertEqual(
            input_params,
            {
                'TargetedWrapperMap': {'foo': {'bar': self.transformed_value}},
                'UntargetedWrapperMap': {'foo': {'bar': self.original_value}},
            },
        )

    def test_transform_nested_list(self):
        input_params = {
            'TargetedWrapperList': [
                [self.original_value, self.original_value]
            ],
            'UntargetedWrapperList': [
                [self.original_value, self.original_value]
            ],
        }

        targeted_list_shape = {
            'TransformMe': {
                'type': 'list',
                'member': {'shape': self.target_shape},
            }
        }

        targeted_wrapper_shape = {
            'TargetedWrapperList': {
                'type': 'list',
                'member': {'shape': 'TransformMe'},
            }
        }

        self.add_shape(targeted_list_shape)
        self.add_input_shape(targeted_wrapper_shape)

        untargeted_list_shape = {
            'LeaveAlone': {'type': 'list', 'member': {'shape': 'String'}}
        }

        untargeted_wrapper_shape = {
            'UntargetedWrapperList': {
                'type': 'list',
                'member': {'shape': 'LeaveAlone'},
            }
        }

        self.add_shape(untargeted_list_shape)
        self.add_input_shape(untargeted_wrapper_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        self.assertEqual(
            input_params,
            {
                'TargetedWrapperList': [
                    [self.transformed_value, self.transformed_value]
                ],
                'UntargetedWrapperList': [
                    [self.original_value, self.original_value]
                ],
            },
        )

    def test_transform_incorrect_type_for_structure(self):
        input_params = {'Structure': 'foo'}

        input_shape = {
            'Structure': {
                'type': 'structure',
                'members': {
                    'TransformMe': {'shape': self.target_shape},
                },
            }
        }

        self.add_input_shape(input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        self.assertEqual(input_params, {'Structure': 'foo'})

    def test_transform_incorrect_type_for_map(self):
        input_params = {'Map': 'foo'}

        input_shape = {
            'Map': {
                'type': 'map',
                'key': {'shape': 'String'},
                'value': {'shape': self.target_shape},
            }
        }

        self.add_input_shape(input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        self.assertEqual(input_params, {'Map': 'foo'})

    def test_transform_incorrect_type_for_list(self):
        input_params = {'List': 'foo'}

        input_shape = {
            'List': {'type': 'list', 'member': {'shape': self.target_shape}}
        }

        self.add_input_shape(input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        self.assertEqual(input_params, {'List': 'foo'})



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestResponseMetadata(BaseAWSCommandParamsTest):
    def test_excludes_response_metadata_when_errors_key_present(self):
        # The ResponseMetadata key should never be present in output. For this
        # test case, we are using an arbitrary command, s3api delete-objects,
        # to ensure generally that the ResponseMetadata does not appear in the
        # output even if the parsed response contains an Errors key.
        self.parsed_response = {
            "ResponseMetadata": {
                "RequestId": "REQUEST-ID",
                "HTTPStatusCode": 200,
                "HTTPHeaders": {
                    "x-amz-id-2": "REQUEST-ID-2",
                    "x-amz-request-id": "REQUEST-ID",
                    "date": "Thu, 13 Apr 2023 15:31:52 GMT",
                    "content-type": "application/xml",
                    "transfer-encoding": "chunked",
                    "server": "AmazonS3",
                    "connection": "close",
                },
                "RetryAttempts": 0,
            },
            "Errors": [
                {
                    "Key": "key",
                    "VersionId": "version-id",
                    "Code": "NoSuchVersion",
                    "Message": "The specified version does not exist.",
                }
            ],
        }
        stdout, _, _ = self.run_cmd(
            [
                "s3api",
                "delete-objects",
                "--bucket",
                "bucket",
                "--delete",
                "Objects=[{Key=key,VersionId=version-id}]",
            ]
        )
        self.assertNotIn("ResponseMetadata", stdout)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.clidriver import AWSCLIEntryPoint
from awscli.testutils import (
    BaseAWSCommandParamsTest,
    FileCreator,
    create_clidriver,
)


class TestArgsResolution(BaseAWSCommandParamsTest):
    def setUp(self):
        super(TestArgsResolution, self).setUp()
        self.files = FileCreator()
        config_contents = '[profile bar]\n' 'region = us-west-2\n'
        self.environ['AWS_CONFIG_FILE'] = self.files.create_file(
            'myconfig', config_contents
        )
        self.driver = create_clidriver()
        self.entry_point = AWSCLIEntryPoint(self.driver)

    def tearDown(self):
        super(TestArgsResolution, self).tearDown()
        self.files.remove_all()

    def test_profile_resolution_order(self):
        self.environ['AWS_PROFILE'] = 'foo'
        self.parsed_responses = [{"Reservations": []}]
        self.run_cmd('--profile bar ec2 describe-instances', expected_rc=0)
        self.assertEqual(self.driver.session.profile, 'bar')

    def test_can_get_version_with_non_existent_profile(self):
        self.environ['AWS_PROFILE'] = 'foo'
        # ProfileNotFound exception shouldn't be raised
        self.run_cmd('--version', expected_rc=0)

    def test_can_get_list_profiles_with_non_existent_profile_in_env_var(self):
        self.environ['AWS_PROFILE'] = 'foo'
        # ProfileNotFound exception shouldn't be raised
        stdout, _, _ = self.run_cmd('configure list-profiles')
        self.assertEqual(stdout, 'bar\n')



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import shutil
import tempfile

from botocore.session import Session

from awscli.autocomplete import db
from awscli.autocomplete.local import indexer, model
from awscli.testutils import mock, unittest

# Quick note about these tests.  sqlite3 is used as the data store for the
# index cache.  When testing this we have two options.  We can either invoke
# various methods from the indexer and the query the generate sqlite DB and
# make sure the records make sense, or we can invoke various methods from the
# indexer and then use our ``ModelIndex`` class to query the index.  The
# ``ModelIndex`` is an abstraction that hides sqlite3 from the clients of the
# index.  I've decided to test the two together, that is, use our indexer and
# the ``ModelIndexer`` to ensure we're getting the right data.  This has the
# downsides that we don't verify anything that we're writing to our sqlite3
# database.  We could be writing inefficient or completely wrong data to our
# DB, but as last the ``ModelIndexer`` understands how to query the data, we'll
# never know.  The reason I think this approach makes sense is that I suspect
# the schema will change as this feature is developed, and it's more important
# to ensure that we can read the index we generate compared to verifying the
# contents of the index.  Once this schema has stabilized we could go in and
# write tests that work at the sqlite3 layer.


class DummyCommand:
    def __init__(self, command_name, subcommand_table=None, arg_table=None):
        self.name = command_name
        if subcommand_table is None:
            subcommand_table = {}
        self.subcommand_table = subcommand_table
        if arg_table is None:
            arg_table = {}
        self.arg_table = arg_table

    def create_help_command(self):
        help_command = mock.Mock()
        service = mock.Mock()
        service.service_model.metadata = {'serviceFullName': 'Amazon Elastic'}
        help_command.command_table = {'ec2': service}
        return help_command


class DummyArg:
    def __init__(
        self,
        name,
        cli_type_name='string',
        nargs=None,
        positional_arg=False,
        required=False,
        help_text='',
    ):
        self.name = name
        self.cli_type_name = cli_type_name
        self.nargs = nargs
        self.positional_arg = positional_arg
        self.required = required
        self.help_text = help_text


class BaseIndexerTest(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock(spec=Session)
        self.aws_command = DummyCommand(
            # The CLIDriver doesn't actually have a ``name`` property,
            # but it should probably be 'aws'.
            command_name=None,
            arg_table={
                'region': DummyArg('region'),
                'endpoint-url': DummyArg('endpoint-url'),
            },
            subcommand_table={
                'ec2': DummyCommand(
                    command_name='ec2',
                    subcommand_table={
                        'describe-instances': DummyCommand(
                            'describe-instances',
                            arg_table={
                                'instance-ids': DummyArg('instance-ids'),
                                'filters': DummyArg(
                                    'filters', 'list', nargs='+'
                                ),
                            },
                        ),
                        'run-instances': DummyCommand('run-instances'),
                    },
                ),
                's3': DummyCommand(
                    command_name='s3',
                    subcommand_table={
                        'list-objects': DummyCommand('list-objects'),
                        'put-object': DummyCommand('put-object'),
                    },
                ),
            },
        )
        # Ideally we just use ':memory:', but that's for
        # a specific sqlite3 connection so we'd have to
        # update our interfaces to return/accept a db connection.
        # I'd prefer not to have db connections as part of the
        # indexing interface.
        self.temp_dir = tempfile.mkdtemp()
        self.tempfile = os.path.join(self.temp_dir, 'temp.db')
        self.db_conn = db.DatabaseConnection(self.tempfile)
        self.indexer = indexer.ModelIndexer(self.db_conn)
        self.query = model.ModelIndex(self.tempfile)

    def tearDown(self):
        self.db_conn.close()
        # TODO: This object is impossible to clean up after properly it should
        # take a db connection instead of a filename so it doesn't have a
        # private copy of the database connection.
        self.query._db_connection.close()
        shutil.rmtree(self.temp_dir)


class TestCanRetrieveCommands(BaseIndexerTest):
    def test_can_retrieve_top_level_commands(self):
        self.indexer.generate_index(self.aws_command)
        self.assertEqual(
            set(self.query.command_names(lineage=['aws'])), set(['ec2', 's3'])
        )

    def test_can_retrieve_operation_names(self):
        self.indexer.generate_index(self.aws_command)
        self.assertEqual(
            set(self.query.command_names(lineage=['aws', 'ec2'])),
            set(['describe-instances', 'run-instances']),
        )

    def test_can_retrieve_global_params(self):
        self.indexer.generate_index(self.aws_command)
        self.assertEqual(
            set(self.query.arg_names(lineage=[], command_name='aws')),
            set(['region', 'endpoint-url']),
        )

    def test_can_retrieve_service_params(self):
        self.indexer.generate_index(self.aws_command)
        self.assertEqual(
            set(
                self.query.arg_names(
                    lineage=['aws', 'ec2'], command_name='describe-instances'
                )
            ),
            set(['instance-ids', 'filters']),
        )

    def test_can_retrieve_correct_commands_when_shadowed(self):
        # Suppose that 's3' had a 'describe-instances' operation.  We
        # should be able to differentiate between this command and the
        # one in ec2 via the ``parent`` arg.
        s3_commands = self.aws_command.subcommand_table['s3'].subcommand_table
        s3_commands['describe-instances'] = DummyCommand('describe-instances')
        self.indexer.generate_index(self.aws_command)
        # The 'ec2' version has several params in its arg table, but this
        # 's3' version should have no params because we didn't add an arg
        # table.
        self.assertEqual(
            set(
                self.query.arg_names(
                    lineage=['aws', 's3'], command_name='describe-instances'
                )
            ),
            set([]),
        )

    def test_empty_list_when_no_args(self):
        # Service commands don't have arguments.
        self.indexer.generate_index(self.aws_command)
        self.assertEqual(
            set(self.query.arg_names(lineage=['aws'], command_name='ec2')),
            set([]),
        )

    def test_empty_list_on_unknown_service(self):
        self.indexer.generate_index(self.aws_command)
        self.assertEqual(
            set(self.query.command_names(lineage=['aws', 'foobar'])),
            set([]),
        )

    def test_can_get_full_service_name(self):
        self.indexer.generate_index(self.aws_command)
        self.assertEqual(
            set(self.query.commands_with_full_name(lineage=['aws'])),
            set([('ec2', 'Amazon Elastic'), ('s3', 'High level S3 commands')]),
        )

    def test_can_get_argument_data(self):
        self.indexer.generate_index(self.aws_command)
        arg_data = self.query.get_argument_data(
            lineage=['aws', 'ec2'],
            command_name='describe-instances',
            arg_name='filters',
        )
        self.assertEqual(
            arg_data,
            model.CLIArgument(
                argname='filters',
                type_name='list',
                command='describe-instances',
                parent='aws.ec2',
                nargs='+',
                positional_arg='0',
                required=False,
            ),
        )


class TestCanCreateModelIndexer(unittest.TestCase):
    def test_can_create_model_indexer(self):
        index = indexer.create_model_indexer('/tmp/a/b/c/d')
        self.assertIsInstance(index, indexer.ModelIndexer)


class TestGeneratesIndex(BaseIndexerTest):
    def tearDown(self):
        self.db_conn.close()
        shutil.rmtree(self.temp_dir)

    def test_generates_indexes_for_tables(self):
        self.indexer.generate_index(self.aws_command)
        index_info = 'SELECT name from pragma_index_info("%s");'
        index = self.db_conn.execute(index_info % 'parent_index').fetchall()
        self.assertEqual([('parent',)], index)
        index = self.db_conn.execute(
            index_info % 'parent_command_index'
        ).fetchall()
        self.assertEqual([('parent',), ('command',)], index)



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import re

from awscli.testutils import FileCreator, aws, unittest


class TestShow(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()
        self.environ = os.environ.copy()
        self.environ['AWS_CONFIG_FILE'] = self.files.create_file(
            'config', ('[default]\n' 'cli_history = enabled')
        )
        self.environ['AWS_DEFAULT_PROFILE'] = 'default'
        self.environ['AWS_DEFAULT_REGION'] = 'us-west-2'
        self.environ['AWS_STS_REGIONAL_ENDPOINTS'] = 'regional'
        self.environ['AWS_CLI_HISTORY_FILE'] = os.path.join(
            self.files.rootdir, 'history.db'
        )

    def tearDown(self):
        self.files.remove_all()

    def remove_color(self, output):
        return re.compile(r'\x1b[^m]*m').sub('', output)

    def assert_contains_in_order(self, lines, contents):
        current_pos = 0
        prev_line = None
        for line in lines:
            self.assertIn(line, contents)
            new_pos = contents.find(line)
            if new_pos < current_pos:
                self.fail(
                    'Line: "%s" should have came after line: "%s"'
                    % (line, prev_line)
                )
            prev_line = line
            current_pos = new_pos

    def test_show(self):
        # Make a call that does not require credentials just in case the
        # user was using the config file to provide credentials.
        cmd = 'sts assume-role-with-saml '
        cmd += '--role-arn  arn:aws:iam::...:invalid '
        cmd += '--principal-arn  arn:aws:iam::...:invalid  '
        cmd += '--saml-assertion fake-assertion'
        aws(cmd, env_vars=self.environ)
        # Now run the show command and make sure the general output is all
        # there.
        result = aws('history show', env_vars=self.environ)
        uncolored_content = self.remove_color(result.stdout)

        self.assert_contains_in_order(
            [
                'AWS CLI command entered',
                'with AWS CLI version: aws-cli/',
                "with arguments: [",
                "'sts', ",
                "'assume-role-with-saml',",
                '[0] API call made',
                'to service: sts',
                'using operation: AssumeRoleWithSAML',
                'with parameters: {',
                '    "PrincipalArn": "arn:aws:iam::...:invalid",',
                '    "RoleArn": "arn:aws:iam::...:invalid",',
                '    "SAMLAssertion": "fake-assertion"',
                '[0] HTTP request sent',
                'to URL: https://sts.us-west-2.amazonaws.com/',
                'with method: POST',
                'with body: ',
                'Action=AssumeRoleWithSAML',
                'Version=2011-06-15',
                '[0] HTTP response received',
                'with status code: 400',
                'with body: <?xml version="1.0" ?>',
                '[0] HTTP response parsed',
                'parsed to: {',
                '    "Error": {',
                'AWS CLI command exited',
                'with return code: 254',
            ],
            uncolored_content,
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import difflib

from awscli.compat import StringIO
from awscli.customizations.datapipeline.listrunsformatter import (
    ListRunsFormatter,
)
from awscli.testutils import mock, unittest


class TestListRunsFormatter(unittest.TestCase):
    def setUp(self):
        self.formatter = ListRunsFormatter(mock.Mock(query=None))
        self.stream = StringIO()

    def assert_data_renders_to(self, data, table):
        self.formatter('list-runs', data, stream=self.stream)
        rendered = self.stream.getvalue()

        differ = difflib.Differ()
        diff = differ.compare(table.splitlines(1), rendered.splitlines(1))

        self.assertEqual(table, rendered, msg='\n' + '\n'.join(diff))

    def test_empty(self):
        self.assert_data_renders_to(
            [],
            "       Name                                                Scheduled Start      Status                 \n"  # noqa
            "       ID                                                  Started              Ended              \n"  # noqa
            "---------------------------------------------------------------------------------------------------\n",
        )  # noqa

    def test_single_row(self):
        self.assert_data_renders_to(
            [
                {
                    '@componentParent': 'parent',
                    '@id': 'id',
                    '@scheduledStartTime': 'now',
                    '@status': 'status',
                    '@actualStartTime': 'actualStartTime',
                    '@actualEndTime': 'actualEndTime',
                }
            ],
            "       Name                                                Scheduled Start      Status                 \n"  # noqa
            "       ID                                                  Started              Ended              \n"  # noqa
            "---------------------------------------------------------------------------------------------------\n"  # noqa
            "   1.  parent                                              now                  status                 \n"  # noqa
            "       id                                                  actualStartTime      actualEndTime      \n"  # noqa
            "\n",
        )



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json

import ruamel.yaml as yaml

from awscli.clidriver import AWSCLIEntryPoint
from awscli.testutils import (
    BaseAWSCommandParamsTest,
    FileCreator,
    create_clidriver,
    if_windows,
    mock,
    skip_if_windows,
)


class TestOutput(BaseAWSCommandParamsTest):
    def setUp(self):
        # Patch this before super, in case called on import
        self.patch_colorama_init = mock.patch('colorama.init')
        self.mock_colorama_init = self.patch_colorama_init.start()

        super().setUp()
        self.files = FileCreator()

        self.patch_popen = mock.patch('awscli.utils.Popen')
        self.mock_popen = self.patch_popen.start()

        self.patch_tty = mock.patch('awscli.utils.is_a_tty')
        self.mock_is_a_tty = self.patch_tty.start()
        self.mock_is_a_tty.return_value = True

        self.cmdline = 'ec2 describe-regions'
        self.parsed_response = {
            "Regions": [
                {
                    "Endpoint": "ec2.ap-south-1.amazonaws.com",
                    "RegionName": "ap-south-1",
                },
            ]
        }
        self.expected_content = self.get_expected_content(self.parsed_response)

    def tearDown(self):
        super().tearDown()
        self.files.remove_all()
        self.patch_popen.stop()
        self.patch_tty.stop()

    def get_expected_content(self, response):
        content = json.dumps(response, indent=4)
        content += '\n'
        return content

    def write_cli_pager_config(self, pager):
        config_file = self.files.create_file(
            'config', '[default]\n' f'cli_pager = {pager}\n'
        )
        self.environ['AWS_CONFIG_FILE'] = config_file
        self.driver = create_clidriver()
        self.entry_point = AWSCLIEntryPoint(self.driver)

    def assert_content_to_pager(
        self, expected_pager, expected_content, expected_less_flags=None
    ):
        actual_pager = self.mock_popen.call_args[1]['args']
        if isinstance(actual_pager, list):
            actual_pager = ' '.join(actual_pager)
        self.assertEqual(expected_pager, actual_pager)
        self.assertEqual(expected_content, self.get_content_sent_to_popen())
        if expected_less_flags:
            actual_env = self.mock_popen.call_args[1]['env']
            self.assertEqual(expected_less_flags, actual_env['LESS'])

    def get_content_sent_to_popen(self):
        content = ''
        popen_stdin = self.mock_popen.return_value.stdin
        for write_call in popen_stdin.write.call_args_list:
            content += write_call[0][0]
        return content

    def test_no_cli_pager_disable_outputs_to_pager(self):
        self.run_cmd(self.cmdline + ' --no-cli-pager')
        self.assertFalse(self.mock_popen.called)

    def test_outputs_to_pager(self):
        self.run_cmd(self.cmdline)
        self.assert_content_to_pager(
            expected_pager=mock.ANY, expected_content=self.expected_content
        )

    def test_does_not_output_to_pager_if_not_tty(self):
        self.mock_is_a_tty.return_value = False
        stdout, _, _ = self.run_cmd(self.cmdline)
        self.assertFalse(self.mock_popen.called)
        self.assertEqual(stdout, self.expected_content)

    @skip_if_windows('less is not used for windows')
    def test_outputs_to_less_non_windows(self):
        self.run_cmd(self.cmdline)
        self.assert_content_to_pager(
            expected_pager='less',
            expected_content=self.expected_content,
            expected_less_flags='FRX',
        )

    @if_windows('more is only used for windows')
    def test_outputs_to_more_for_windows(self):
        self.run_cmd(self.cmdline)
        self.assert_content_to_pager(
            expected_pager='more',
            expected_content=self.expected_content,
        )

    def test_respects_aws_pager_env_var(self):
        self.environ['AWS_PAGER'] = 'mypager'
        self.run_cmd(self.cmdline)
        self.assert_content_to_pager(
            expected_pager='mypager',
            expected_content=self.expected_content,
        )

    def test_respects_cli_pager_config_var(self):
        self.write_cli_pager_config('mypager')
        self.run_cmd(self.cmdline)
        self.assert_content_to_pager(
            expected_pager='mypager',
            expected_content=self.expected_content,
        )

    def test_respects_pager_env_var(self):
        self.environ['PAGER'] = 'mypager'
        self.run_cmd(self.cmdline)
        self.assert_content_to_pager(
            expected_pager='mypager',
            expected_content=self.expected_content,
        )

    def test_no_pager_if_configured_to_empty_str(self):
        self.environ['AWS_PAGER'] = ''
        stdout, _, _ = self.run_cmd(self.cmdline)
        self.assertFalse(self.mock_popen.called)
        self.assertEqual(stdout, self.expected_content)

    def test_aws_pager_env_var_beats_cli_pager_config_var(self):
        self.environ['AWS_PAGER'] = 'envpager'
        self.write_cli_pager_config('configpager')
        self.run_cmd(self.cmdline)
        self.assert_content_to_pager(
            expected_pager='envpager',
            expected_content=self.expected_content,
        )

    def test_aws_pager_env_var_beats_pager_env_var(self):
        self.write_cli_pager_config('configpager')
        self.environ['PAGER'] = 'envpager'
        self.run_cmd(self.cmdline)
        self.assert_content_to_pager(
            expected_pager='configpager',
            expected_content=self.expected_content,
        )

    def test_cli_pager_config_var_beats_pager_env_var(self):
        self.environ['AWS_PAGER'] = 'awspager'
        self.environ['PAGER'] = 'envpager'
        self.run_cmd(self.cmdline)
        self.assert_content_to_pager(
            expected_pager='awspager',
            expected_content=self.expected_content,
        )

    def test_respects_less_env_var(self):
        self.environ['AWS_PAGER'] = 'less'
        self.environ['LESS'] = 'S'
        self.run_cmd(self.cmdline)
        self.assert_content_to_pager(
            expected_pager='less',
            expected_content=self.expected_content,
            expected_less_flags='S',
        )

    def test_raises_exception_when_pager_cannot_be_opened(self):
        self.environ['AWS_PAGER'] = 'no-exists'
        pager_error_message = 'Cannot open no-exists'
        self.mock_popen.side_effect = FileNotFoundError(pager_error_message)
        _, stderr, _ = self.run_cmd(self.cmdline, expected_rc=253)
        self.assertIn('Unable to redirect output to pager', stderr)
        self.assertIn(pager_error_message, stderr)

    def test_colorama_init_not_called(self):
        """
        For https://github.com/aws/aws-cli/issues/9864,
        we had an issue where colorama was always being initialized, instead of
        just during the custom commands or modes that use it
        """
        self.run_cmd(self.cmdline)
        self.mock_colorama_init.assert_not_called()


class TestYAMLStream(BaseAWSCommandParamsTest):
    def assert_yaml_response_equal(self, response, expected):
        with self.assertRaises(ValueError):
            json.loads(response)
        loaded = self.yaml.load(response)
        self.assertEqual(loaded, expected)

    def test_yaml_stream_single_response(self):
        cmdline = 'dynamodb list-tables --output yaml-stream --no-paginate'
        self.parsed_responses = [{'TableNames': ['MyTable']}]
        stdout, _, _ = self.run_cmd(cmdline)
        self.assert_yaml_response_equal(stdout, [{'TableNames': ['MyTable']}])

    def test_yaml_stream_paginated_response(self):
        cmdline = 'dynamodb list-tables --output yaml-stream'
        self.parsed_responses = [
            {'TableNames': ['MyTable'], 'LastEvaluatedTableName': 'MyTable'},
            {'TableNames': ['MyTable2']},
        ]
        stdout, _, _ = self.run_cmd(cmdline)
        self.assert_yaml_response_equal(
            stdout,
            [
                {
                    'TableNames': ['MyTable'],
                    'LastEvaluatedTableName': 'MyTable',
                },
                {'TableNames': ['MyTable2']},
            ],
        )

    def test_yaml_stream_removes_response_metadata(self):
        cmdline = 'dynamodb list-tables --output yaml-stream --no-paginate'
        self.parsed_responses = [
            {
                'TableNames': ['MyTable'],
                'ResponseMetadata': {'RequestId': 'id'},
            }
        ]
        stdout, _, _ = self.run_cmd(cmdline)
        self.assert_yaml_response_equal(stdout, [{'TableNames': ['MyTable']}])

    def test_yaml_stream_removes_response_metadata_for_all_responses(self):
        cmdline = 'dynamodb list-tables --output yaml-stream'
        self.parsed_responses = [
            {
                'TableNames': ['MyTable'],
                'LastEvaluatedTableName': 'MyTable',
                'ResponseMetadata': {'RequestId': 'id'},
            },
            {
                'TableNames': ['MyTable2'],
                'ResponseMetadata': {'RequestId': 'id2'},
            },
        ]
        stdout, _, _ = self.run_cmd(cmdline)
        self.assert_yaml_response_equal(
            stdout,
            [
                {
                    'TableNames': ['MyTable'],
                    'LastEvaluatedTableName': 'MyTable',
                },
                {'TableNames': ['MyTable2']},
            ],
        )

    def test_yaml_stream_uses_query(self):
        cmdline = (
            'dynamodb list-tables --output yaml-stream --no-paginate '
            '--query TableNames'
        )
        self.parsed_responses = [{'TableNames': ['MyTable']}]
        stdout, _, _ = self.run_cmd(cmdline)
        self.assert_yaml_response_equal(stdout, [['MyTable']])

    def test_yaml_stream_uses_query_across_each_response(self):
        cmdline = (
            'dynamodb list-tables --output yaml-stream --query TableNames'
        )
        self.parsed_responses = [
            {'TableNames': ['MyTable'], 'LastEvaluatedTableName': 'MyTable'},
            {'TableNames': ['MyTable2']},
        ]
        stdout, _, _ = self.run_cmd(cmdline)
        self.assert_yaml_response_equal(stdout, [['MyTable'], ['MyTable2']])

    def test_yaml_stream_with_empty_response(self):
        cmdline = 's3api delete-bucket --bucket mybucket --output yaml-stream'
        self.parsed_responses = [{}]
        stdout, _, _ = self.run_cmd(cmdline)
        self.assertEqual(stdout, '')



#!/usr/bin/env python
# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

from awscli.testutils import BaseAWSCommandParamsTest, skip_if_windows


class TestCreateVirtualMFADevice(BaseAWSCommandParamsTest):
    prefix = 'iam create-virtual-mfa-device'

    def setUp(self):
        super(TestCreateVirtualMFADevice, self).setUp()
        self.parsed_response = {
            'ResponseMetadata': {
                'HTTPStatusCode': 200,
                'RequestId': 'requset-id',
            },
            "VirtualMFADevice": {
                "Base32StringSeed": (
                    "VFpYTVc2V1lIUFlFRFczSVhLUlpRUTJRVFdUSFRNRDNTQ0c3"
                    "TkZDUVdQWDVETlNWM0IyUENaQVpWTEpQTlBOTA=="
                ),
                "SerialNumber": "arn:aws:iam::419278470775:mfa/fiebaz",
                "QRCodePNG": (
                    "iVBORw0KGgoAAAANSUhEUgAAAPoAAAD6CAIAAAAHjs1qAAAFi"
                    "ElEQVR42u3bQW7jMBAEwPz/07vHvS0QeLpnZFVf7cgyWRTgJv"
                    "PzR+Q1+TEEgrsI7iK4i+AugrsI7iK4i+AugrsI7iK4C+4iuIv"
                    "gLoK7CO4iuIvgLoK7CO4iBe4/rfz/c391k7lLffK5v/r6tZu8"
                    "Ofu444477rjjjjvuuOOOO+6xwcoZ/WTl5D53cNXlZqG2VPpXx"
                    "h133HHHHXfccccdd9xxD09/rU7ZylZdVnvo5BY/7rjjjjvuuO"
                    "OOO+64447713H/RFIOZW0944477rjjjjvuuOOOO+64417ZVM8"
                    "ZPbLHXiOLO+6444477rjjjjvuuOP+fO5bC2lwMeQKoiN/ew0l"
                    "7rjjjjvuuOOOO+644457uCLYGmivHuxeBmcfd6/ijrtXccfdq"
                    "7jj7lXcH5han3Bkeedana9SgTvuuOOOO+6444477ri/hXttc7"
                    "umMLc2tm5ycI4GnzJfWETijjvuuOOOO+6444477s9oZnKDlZv"
                    "vGp2tnf/BNx8/NYA77rjjjjvuuOOOO+64v5V7rZkZHKzalXOw"
                    "cg3J1qorFGK444477rjjjjvuuOOO+1u5fwLrSG9T6zEGxyr3F"
                    "WqPwkeemcEdd9xxxx133HHHHXfcj3Kv4chNUo17rU7ZmtDjax"
                    "J33HHHHXfccccdd9xxfw33rWYm12McOY/wBQupZgN33HHHHXf"
                    "ccccdd9xxx/2DwTpy0Pzm2sh9oyOgc+sZd9xxxx133HHHHXfc"
                    "ccc9liOVSE3SkYIot2KPFES444477rjjjjvuuOOO+/u4DzYzg"
                    "xoGpyH35q3R2Nrb758awB133HHHHXfccccdd9zfyj23NmpX3j"
                    "omkBuc3N7+4DIr+MYdd9xxxx133HHHHXfc38q91gkcUZi7jcG"
                    "d/9xIHnkG4Y477rjjjjvuuOOOO+6430gOR20hbVUTW4cIrg0O"
                    "7rjjjjvuuOOOO+644/4a7keKi8G+6At273Nv3vpc3HHHHXfcc"
                    "ccdd9xxxx332LjXyocanUfgqJ0pqNU4uOOOO+6444477rjjjj"
                    "vuS13E1l53jc5Ndjf144477rjjjjvuuOOOO+64x77Skf352mZ"
                    "+7VhEbo62DiBE6jvccccdd9xxxx133HHHHffs9A8WNbnprz0a"
                    "tv7rYPDrP/IQAe6444477rjjjjvuuON+lPvguK//ci+TrS3Rw"
                    "Z7q5nMEd9xxxx133HHHHXfccce9VSDkippcY5A7nrBVaxyp2n"
                    "DHHXfccccdd9xxxx133FuTVFs5tUttlVq5pZJ7cCy0arjjjjv"
                    "uuOOOO+6444477tmd8K0SIDdnNXZHirj+YwV33HHHHXfccccd"
                    "d9xxfw33mu+bHUiObK4By7E7jgF33HHHHXfccccdd9xxfw33W"
                    "hcx+Kt/a1Zyn5tbZltkcccdd9xxxx133HHHHXfcb4DOnRp4Yg"
                    "eSQ/l9jRDuuOOOO+6444477rjjjntM4WCBMGhlcHByt1GrcbZ"
                    "OSeCOO+6444477rjjjjvuuD8htUMENcFbN1mrcXIfhDvuuOOO"
                    "O+6444477rjjvvSrP3cbg3NWG7pcMTV4OuPtRSTuuOOOO+644"
                    "4477rjjfoV7rXvZWjlbVr6+5CkUNbjjjjvuuOOOO+6444477q"
                    "sacn+7VeMcaYRyVdsjDxHgjjvuuOOOO+6444477rinNNR6myf"
                    "exlbXpJnBHXfccccdd9xxxx133J/PvbAFPT5nN4upLQy44447"
                    "7rjjjjvuuOOOO+4t7rWx27pybY/9iLMjvQ3uuOOOO+6444477"
                    "rjjjvvSfB/Zr86t58H53qpxBpdoYZxxxx133HHHHXfccccd99"
                    "dwF3lKcBfcRXAXwV0EdxHcRXAXwV0EdxHcRXAXwV1wF8FdBHc"
                    "R3EVwF8FdBHcR3EVwF8Fd5F/+AgASajf850wfAAAAAElFTkSu"
                    "QmCC"
                ),
            },
        }

    def getpath(self, filename):
        return os.path.join(
            os.path.abspath(os.path.dirname(__file__)), filename
        )

    def remove_file_if_exists(self, filename):
        if os.path.isfile(filename):
            os.remove(filename)

    def test_base32(self):
        outfile = self.getpath('fiebaz.b32')
        self.addCleanup(self.remove_file_if_exists, outfile)
        cmdline = self.prefix
        cmdline += ' --virtual-mfa-device-name fiebaz'
        cmdline += (
            ' --outfile %s --bootstrap-method Base32StringSeed' % outfile
        )
        result = {"VirtualMFADeviceName": 'fiebaz'}
        self.assert_params_for_cmd(cmdline, result)
        self.assertTrue(os.path.exists(outfile))

    def test_qrcode(self):
        outfile = self.getpath('fiebaz.png')
        self.addCleanup(self.remove_file_if_exists, outfile)
        cmdline = self.prefix
        cmdline += ' --virtual-mfa-device-name fiebaz'
        cmdline += ' --outfile %s --bootstrap-method QRCodePNG' % outfile
        result = {"VirtualMFADeviceName": 'fiebaz'}
        self.assert_params_for_cmd(cmdline, result)
        self.assertTrue(os.path.exists(outfile))

    def test_bad_filename(self):
        outfile = '/some/bad/filename.png'
        self.addCleanup(self.remove_file_if_exists, outfile)
        cmdline = self.prefix
        cmdline += ' --virtual-mfa-device-name fiebaz'
        cmdline += ' --outfile %s --bootstrap-method QRCodePNG' % outfile
        self.assert_params_for_cmd(cmdline, expected_rc=252)

    def test_relative_filename(self):
        outfile = 'filename.png'
        self.addCleanup(self.remove_file_if_exists, outfile)
        cmdline = self.prefix
        cmdline += ' --virtual-mfa-device-name fiebaz'
        cmdline += ' --outfile %s --bootstrap-method QRCodePNG' % outfile
        result = {"VirtualMFADeviceName": 'fiebaz'}
        self.assert_params_for_cmd(cmdline, result)
        self.assertTrue(os.path.exists(outfile))

    def test_bad_relative_filename(self):
        outfile = 'some/bad/filename.png'
        self.addCleanup(self.remove_file_if_exists, outfile)
        cmdline = self.prefix
        cmdline += ' --virtual-mfa-device-name fiebaz'
        cmdline += ' --outfile %s --bootstrap-method QRCodePNG' % outfile
        self.assert_params_for_cmd(cmdline, expected_rc=252)

    def test_bad_response(self):
        # This can happen if you run the create-virtual-mfa-device
        # command multiple times with the same name.  You'll get
        # an "already exists" error and we should handle that case
        # gracefully.
        self.parsed_response = {
            'Error': {
                'Code': 'EntityAlreadyExists',
                'Message': 'MFADevice entity at the and name already exists.',
                'Type': 'Sender',
            },
            'ResponseMetadata': {
                'HTTPStatusCode': 409,
                'RequestId': 'requset-id',
            },
        }
        self.http_response.status_code = 409
        cmdline = self.prefix
        cmdline += ' --virtual-mfa-device-name fiebaz'
        cmdline += ' --outfile foo --bootstrap-method QRCodePNG'
        # The error message should be in the stderr.
        self.assert_params_for_cmd(
            cmdline,
            stderr_contains=self.parsed_response['Error']['Message'],
            expected_rc=254,
        )

    @skip_if_windows("Permissions test not valid on Windows.")
    def test_output_file_permissions(self):
        outfile = self.getpath('fiebaz_perms.b32')
        self.addCleanup(self.remove_file_if_exists, outfile)
        cmdline = self.prefix
        cmdline += ' --virtual-mfa-device-name fiebaz'
        cmdline += (
            ' --outfile %s --bootstrap-method Base32StringSeed' % outfile
        )
        result = {"VirtualMFADeviceName": 'fiebaz'}
        self.assert_params_for_cmd(cmdline, result)
        self.assertEqual(os.stat(outfile).st_mode & 0xFFF, 0o600)

    @skip_if_windows("Permissions test not valid on Windows.")
    def test_output_file_permissions_existing_file(self):
        outfile = self.getpath('fiebaz_perms_existing.b32')
        self.addCleanup(self.remove_file_if_exists, outfile)
        with open(outfile, 'wb') as f:
            f.write(b'existing')
        os.chmod(outfile, 0o644)
        cmdline = self.prefix
        cmdline += ' --virtual-mfa-device-name fiebaz'
        cmdline += (
            ' --outfile %s --bootstrap-method Base32StringSeed' % outfile
        )
        result = {"VirtualMFADeviceName": 'fiebaz'}
        self.assert_params_for_cmd(cmdline, result)
        self.assertEqual(os.stat(outfile).st_mode & 0xFFF, 0o600)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json

from awscli.testutils import BaseAWSCommandParamsTest, FileCreator


class TestDeployCommand(BaseAWSCommandParamsTest):
    def setUp(self):
        super(TestDeployCommand, self).setUp()
        self.files = FileCreator()
        self.parsed_responses = [
            # First it checks to see if a stack with that name exists. So
            # we fake a response indicating that the stack exists and is in
            # an OK state.
            {
                'Stacks': {
                    'StackName': 'Stack',
                    'StackStatus': 'UPDATE_COMPLETE',
                }
            },
            # Now it creates a changeset, so we fake a response with an ID.
            {'Id': 'FakeChangeSetId'},
            # This fakes a failed response from the waiter because the
            # changeset was empty.
            {
                'StackName': 'Stack',
                'Status': 'FAILED',
                'StatusReason': (
                    'The submitted information didn\'t contain changes. '
                    'Submit different information to create a change set.'
                ),
                'ExecutionStatus': 'UNAVAILABLE',
            },
        ]
        # The template is inspected before we make any of the calls so it
        # needs to have valid JSON content.
        path = self.files.create_file('template.json', '{}')
        self.command = (
            'cloudformation deploy --template-file %s ' '--stack-name Stack'
        ) % path

    def tearDown(self):
        self.files.remove_all()
        super(TestDeployCommand, self).tearDown()

    def test_does_return_zero_exit_code_on_empty_changeset_by_default(self):
        self.run_cmd(self.command, expected_rc=0)

    def test_does_return_zero_exit_code_on_empty_changeset(self):
        self.command += ' --no-fail-on-empty-changeset'
        self.run_cmd(self.command, expected_rc=0)

    def test_does_return_non_zero_exit_code_on_empty_changeset(self):
        self.command += ' --fail-on-empty-changeset'
        self.run_cmd(self.command, expected_rc=255)


class TestDeployCommandParameterOverrides(TestDeployCommand):
    def setUp(self):
        super(TestDeployCommandParameterOverrides, self).setUp()
        template = '''{
          "AWSTemplateFormatVersion": "2010-09-09",
          "Parameters": {
            "Key1": {
              "Type": "String"
            },
            "Key2": {
              "Type": "String"
            }
          }
        }'''
        path = self.files.create_file('template.json', template)
        self.command = (
            'cloudformation deploy --template-file %s ' '--stack-name Stack '
        ) % path

    def _assert_parameters_parsed(self):
        self.assertEqual(
            self.operations_called[1][1]['Parameters'],
            [
                {'ParameterKey': 'Key1', 'ParameterValue': 'Value1'},
                {'ParameterKey': 'Key2', 'ParameterValue': 'Value2'},
            ],
        )

    def create_json_file(self, filename, data):
        return self.files.create_file(filename, json.dumps(data))

    def test_parameter_overrides_shorthand(self):
        self.command += ' --parameter-overrides Key1=Value1 Key2=Value2'
        self.run_cmd(self.command)
        self._assert_parameters_parsed()

    def test_parameter_overrides_from_inline_original_json(self):
        original_like_json = ['Key1=Value1', 'Key2=Value2']
        path = self.create_json_file('param.json', original_like_json)
        self.command += ' --parameter-overrides file://%s' % path
        self.run_cmd(self.command)
        self._assert_parameters_parsed()

    def test_parameter_overrides_from_inline_cf_like_json(self):
        cf_like_json = (
            '[{"ParameterKey":"Key1",'
            '"ParameterValue":"Value1"},'
            '{"ParameterKey":"Key2",'
            '"ParameterValue":"Value2"}]'
        )
        self.command += ' --parameter-overrides %s' % cf_like_json
        self.run_cmd(self.command)
        self._assert_parameters_parsed()

    def test_parameter_overrides_from_cf_like_json_file(self):
        cf_like_json = [
            {'ParameterKey': 'Key1', 'ParameterValue': 'Value1'},
            {'ParameterKey': 'Key2', 'ParameterValue': 'Value2'},
        ]
        path = self.create_json_file('param.json', cf_like_json)
        self.command += ' --parameter-overrides file://%s' % path
        self.run_cmd(self.command)
        self._assert_parameters_parsed()

    def test_parameter_overrides_from_inline_codepipeline_like_json(self):
        codepipeline_like_json = (
            '{"Parameters":{"Key1":"Value1",' '"Key2":"Value2"}}'
        )
        self.command += ' --parameter-overrides %s' % codepipeline_like_json
        self.run_cmd(self.command)
        self._assert_parameters_parsed()

    def test_parameter_overrides_from_codepipeline_like_json_file(self):
        codepipeline_like_json = {
            'Parameters': {'Key1': 'Value1', 'Key2': 'Value2'}
        }
        path = self.create_json_file('param.json', codepipeline_like_json)
        self.command += ' --parameter-overrides file://%s' % path
        self.run_cmd(self.command)
        self._assert_parameters_parsed()

    def test_parameter_overrides_from_original_json_file(self):
        original_like_json = ['Key1=Value1', 'Key2=Value2']
        path = self.create_json_file('param.json', original_like_json)
        self.command += ' --parameter-overrides file://%s' % path
        self.run_cmd(self.command, expected_rc=0)
        self._assert_parameters_parsed()

    def test_parameter_overrides_from_invalid_cf_like_json_file(self):
        invalid_cf_like_json = [
            {
                'ParameterKey': 'Key1',
                'ParameterValue': 'Value1',
                'RedundantKey': 'RedundantValue',
            },
            {'ParameterKey': 'Key2', 'ParameterValue': 'Value2'},
        ]
        path = self.create_json_file('param.json', invalid_cf_like_json)
        self.command += ' --parameter-overrides file://%s' % path
        _, err, _ = self.run_cmd(self.command, expected_rc=252)
        self.assertTrue('JSON passed to --parameter-overrides must be' in err)

    def test_parameter_overrides_from_invalid_json(self):
        cf_like_json = {'SomeKey': [{'RedundantKey': 'RedundantValue'}]}
        path = self.create_json_file('param.json', cf_like_json)
        self.command += ' --parameter-overrides file://%s' % path
        _, err, _ = self.run_cmd(self.command, expected_rc=252)
        self.assertTrue('JSON passed to --parameter-overrides must be' in err)
